(*p

\setlength{\parindent}{0cm}
\setlength{\parskip}{24pt} % marche pas!

*)

(*s Analyzing messages *)

open Dns
open Answer

(* 

As we saw in the Dns module, parsing DNS messages to get a list of
RRsets for each section is not as easy as it should be.  The next step
is interpreting the information conveyed, and it's again a bit
cumbersome at least.  Cnames and negative answers add quite a bit of
what seems mainly gratuitous complexity.  There's one entry point in
this module: the [process_answer] function.  Its duty is to analyze
the answer message, using the query context.  The outcome is a value
of type [answer_t] (excluding the [RFailed] case, that should be used
only as final resolution result) along with a list of cnames.


*)


(*s First checks *)

(* The first thing we'll have to do is perform some simple checks on
   the message.  In some cases, we can discard most of the message
   content and return directly our result:
   
   \begin{itemize}
   \item the message is invalid (something's wrong with its header or
     source),
   \item it is an error message (rcode not 0 nor 3),  
   \item it is truncated.
   \end{itemize}
   
   As a trick, this function returns [Error 0] to indicate the message
   is ok and we'll have to analyze it for real.  *)

let first_checks ip id sname stype sclass m = 

(* Basic checks.  I don't think we need to check the source port is
   indeed 53. The destination (local) port is checked by the socket
   itself.
   
   The answer should not have the RD bit (since we don't set it in
   our queries and it is to be copied from the query rfc1035-4.1.1),
   but some servers do set it so we just don't look at it...  *)
  
  if m.opcode <> 0  then Invalid "bad opcode"   else
  if not m.qr       then Invalid "no QR bit"    else
(*c  if m.rd           then Invalid "RD bit"       else*) (* TODO: log it? *)
  if m.id  <> id    then Invalid "bad id"       else
  if m.adr <> ip    then Invalid "bad source"   else
  if m.tc           then Truncated              else 

(* If the rcode is neither 0 nor 3, there's no point in looking at
   the rest of the message.  *)

  if m.rcode<>0 && m.rcode<>3 then Error(m.rcode) else

(* Only now can we check the question, since error messages
   sometimes have empty questions.  In particular, FORMERR answers
   due to unsupported EDNS sometimes have an empty question.  This
   is in contradiction with rfc5452-4.2. *)
  
  if m.que <> [ sname, stype, sclass ] then Invalid "bad question" else

(* Admittedly strange way to say all is ok! *)
  Error 0

(*s Building an [smsg] *)

(* Once we know we should analyze the message we'll start by turning
it into a simplified from, of type [smsg].  An [smsg] is simple because:

\begin{itemize}
\item it sorts out the cname muddle, by having a section [an]
containing answer rrsets proper, and a section [cn] containing a correct chain
of cnames, if any,
\item it is free of out-of-bailywick rrsets.
\end{itemize}

Additionaly we'll refuse to build an [smsg] if the answer section
contains rrsets of a type other than qtype (except for cnames and
qtype any, of course).

[nxd] is set if the rcode is equal to 3, otherwise it is 0 (we won't
build a [smsg] for a message with another rcode value).

[aab] is the AA bit.

 *)

type smsg = 
    { nxd:bool; aab:bool; qn:name; cn:rrset_t list; 
      an:rrset_t list; au:rrset_t list; ad:rrset_t list;
      ou:rrset_t list }



(* To build a [smsg] we need to separate cnames from answer rrsets
   proper.  The following function does that, from the search name and
   the content of the answer section.  It returns three informations:

\begin{itemize}  
\item the new qname,
\item the list of cname RRs,
\item the list of other (answer proper) RRsets.
\end{itemize}

The qname is defined by rfc2308-1 as 

\begin{quote}
   the name in the query section of an answer, or where this
   resolves to a CNAME, or CNAME chain, the data field of the last
   CNAME
\end{quote}

While extracting cnames, it is checked they form a correct chaining.
An exception is raised if they don't.  We don't require them to be in
proper order.  Let's have a laugh at this other excerpt from
rfc2308-1:

\begin{quote} 
   Implementations should note that including CNAME records in
   responses in order, so that the first has the label from the query
   section, and then each in sequence has the label from the data
   section of the previous (where more than one CNAME is needed)
   allows the sequence to be processed in one pass, and considerably
   eases the task of the receiver.
\end{quote}
   
 *)

let analyze_ans_section sname answ =
  let cnames, others = 
    List.partition (fun (_, _, d) -> rrt_of_rdset_t d = rrt_cname) answ
  in
(* Check the chaining of CNAMEs and get the qname *)
  let rec loop qname rrs =
    if rrs = [] then qname
    else 
      let [ (n,t,d) ], rest =
	List.partition (fun (n, _, _) -> n = qname) rrs
      in
      let CNAME n' = d in
      loop n' rest
  in
  loop sname cnames, cnames, others

(* Check that list of rrsets [an] (intended to be [an] field of an
   [smsg]) contains only rrsets of type [stype].  If we were to
   support DNSSEC, this test wouldn't make sense.
  *)

let check_an stype an =
  if stype=rrt_any then true
  else
  List.filter (fun (n,t,rr) -> rrt_of_rdset_t rr <> stype) an = []
  
(*

Another nice property of [smsg]s is that they won't contain any rrset
the sender has no authority on.  rfc5452-6 requires we filter out
out-of-bailywick rrsets for security reasons.

*)

let in_bailywick n zone = is_indomain n zone

let filter_out_nb_rrsets zone rrsets =
  List.partition (fun (n, _, _) -> in_bailywick n zone) rrsets


(* Build an [smsg].  This can fail for two kinds of problem with the
   original answer section:
   \begin{itemize}
   \item bad cname chaining,
   \item rrsets not of the queried type.
   \end{itemize}
   
   We return a value of type [choice], [Good] being used for a
   positive result, [Bad] for a failure.  *)

type ('a,'b) choice = Good of 'a | Bad of 'b

(*
   If [stype] is [rrt_cname], we check there's no more than one cname,
   and put it in the [an] section.

  *)

let build_smsg sname stype zone m =

(* Filter out rrsets not in bailywick for this server.  Have you
   noticed lots of rrsets in rfc2308-2.1 examples should be filtered
   out?  *)

  let (answ,o1), (aut,o2), (add,o3) =
    let f = filter_out_nb_rrsets zone in
    f m.ans, f m.aut, f m.add
  in
  let ou = o1 @ o2 @ o3 in

(* Build simplified message.  Catch exception for bad cname
   chaining. *)

  try
    let qname, cnames, ans = analyze_ans_section sname answ
    in
    if not (check_an stype ans) then Bad "Invalid rrt in 'an' section"
    else
      if stype = rrt_cname then
	if List.length cnames < 2 then
	  Good { nxd = m.rcode=3; aab=m.aa; qn=qname; 
		 cn=[]; an=cnames; au=aut; ad=add; ou=ou }
	else Bad "Several cnames, but stype = rrt_cname"
      else
	Good { nxd = m.rcode=3; aab=m.aa; qn=qname; 
	       cn=cnames; an=ans; au=aut; ad=add; ou=ou }
  with _ -> 
    Bad "Bad cname chaining"



(*s Analyzing an [smsg] *)

(* 

Once we have a [smsg], we can determine what kind of answer it is.
Either the server has authority on the question name, so the aa bit is
set, and:

\begin{itemize}
\item rcode=3, answer is [NxDomain]
\item one or more rrsets in [an] section, it's [Data]
\item empty [cn] and [an] section, it's [NoData]
\item empty [an] section, non-empty [cn], it's an [Alias].
\end{itemize}

If the aa bit is not set, then either it's a referral or the server is
lame (it is supposed to have authority on this name but has not).
Lame servers generally answer by a referral either on a domain that is
not below the one they're supposed to have authority on (often the
root domain) or to this exact domain.  Note however that we previously
filtered rrsets not under bailywick so the aut section would appear
empty in the first case.

Thus, a valid referral does not have the aa bit, but does include an
NS RRset in the aut section, and it is not an ``auto referral''.

As usual, there's more to it.  Some servers do send authoritative
answers \textbf{without} the aa bit.  They're not lame, just
mind-blowingly stupids.  We know examples of such servers that return
a cname.

*)

(*

  Extract NS rrsets from authority section of message [m].
  ZZZ should check only one?
  *)

let get_ns_rrsets m =
  try
    let n,t,d = List.find (fun (n,t,d) -> rrt_of_rdset_t d = rrt_ns) m.au
    in Some(n, t, d)
  with Not_found -> None


(* 

Negative answers ([NoData] and [NxDomain]) should provide a SOA RR in
the auth section, used to carry the TTL of the negative information.

The following function gets the ttl for a negative answer, along with
the SOA domain name.  The searched ttl is the ttl of the SOA RR in the
authority section, if any, otherwise we shouldn't cache the negative
answer, so we return 0 as ttl.

*)

let get_neg_ttl m = 
  try
    let n, ttl, _ =
      List.find (fun (n, _, rr) -> rrt_of_rdset_t rr = rrt_soa) m.au
    in n, ttl
  with Not_found -> root_name, 0

(* The [analyze_smsg] function itself.  It implements what we
   described above. *)

let analyze_smsg stype zone m =

  if m.aab then
    if m.nxd then
      let n,t = get_neg_ttl m in NxDomain(t,n)
    else
      let lan = List.length m.an
      and lcn = List.length m.cn
      in
      if lan>1 then 
	if stype = rrt_any then 
	  DataAny(List.map (fun (n,t,d) -> t,d) m.an)
	else Invalid("several rrsets in 'an' section")
      else if lan=1 then
	let [n,t,d] = m.an in Data(t,d)

      (* Section [an] is empty *)

      else if lcn = 0 then 
	let n,t = get_neg_ttl m in NoData(t,n)
      else
	Alias m.qn

  else (* aa bit not set *)
    if m.nxd then Invalid "nxd without aa bit"
    else
      match get_ns_rrsets m with
      | Some (n, ttl, rr) when is_subdomain n zone ->
	  Referral(m.qn,  (n,ttl,rr), (m.ad @ m.ou))

      | _ -> 
	  (* Some stupid servers answer a CNAME without the aa bit. *)
	  if List.length m.cn = 0 then Lame root_name
	  else Alias m.qn



(*s Message processing *)

(* Here's our ``main'' function.  Processing an answer message
involves several steps:
\begin{enumerate}
\item check the message warrants more processing,
\item simplify the [msg] to a [smsg], then it's a valid non error message,
\item analyze the answer,
\item add information to the cache.
\end{enumerate}

Here [zone] is the zone the server has (is supposed to have) authority on. *)

let process_answer ip id sname stype sclass zone m = (* ZZZ + AA ? *)

(* First basic checks *)
  let r = first_checks ip id sname stype sclass m in
  if r <> Error 0 then
    [], r
  else

(* Passed, build [smsg] *)
  match build_smsg sname stype zone m with 
  | Bad s -> [], Invalid s
  | Good m ->

(* Succeeded, analyze it *)

  match analyze_smsg stype zone m
  with Invalid i -> [], Invalid i
  | Lame n -> dbg "lame server"; [], Lame n
  | res ->    m.cn, res
