open Dns
open Dns_txt
open Answer
module Misc = Muthr_misc

(*s Basic things *)

let dbg_name n t =
  dbg (Printf.sprintf "C: %s %s" (string_of_name n) t)

let hash_keys hash =
  Hashtbl.fold (fun k v acc -> k :: acc) hash []

type date = float

(* Convert ttl to/from absolute time *)
(*ZZZ [int_of_float] truncates !!! *)

let ttl_of t = int_of_float (t -. (Unix.gettimeofday ()))
let ttl_of t = 
  let at = Unix.gettimeofday () in
  if at > t then -1 else int_of_float (t -. at)

let of_ttl t = (float_of_int t +. (Unix.gettimeofday ()))

(* We raise Expired when a found ttl is -1, it means the cache entry
  is stale.  *)

exception Expired

let check_ttl_of t =
  let r = ttl_of t in
  if r = -1 then raise Expired else r

(*s The RRset cache *)

(*

  A cache entry is associated to a domain name.  It is either:
  \begin{itemize}
  \item[NxDomain] with associated expiration date, or
  \item[Domain] with an assoc list indexing [rrt]s into [date] * [rdset_t]s,
  \end{itemize}
 *)

type cache_entry =
  | NxDomain of name * date
  | Domain of (rrt * (date * rdset_t)) list

(* The cache is a hash-table of [cache_entry]s.  *)

let cache = Hashtbl.create 25000

let nb_rrsets e =
  match e with
  | NxDomain _ -> 1
  | Domain vs  -> List.length vs

let cache_reset () =
  Hashtbl.clear cache

let cache_get name =
  Hashtbl.find cache name


let purge_entry name =
  let e = cache_get name in
  let ct = Unix.gettimeofday ()
  in
  match e with 
  | NxDomain(n, time) -> if time < ct then (Hashtbl.remove cache name; 1) else 0
  | Domain l ->
      (* filter expired elements from list *)
      let l' = List.filter
	  (fun (k,(t,v)) -> t >= ct)
	  l
      in
      Hashtbl.remove cache name;
      if l' <> [] then Hashtbl.add cache name (Domain l');
      List.length l - List.length l'

let purge () =
  let purge_and_count (n,e) name =
    let nr = purge_entry name in (n+if nr>0 then 1 else 0), e+nr
  in
  List.fold_left purge_and_count (0,0) (hash_keys cache)


(* Return the rrset associated to name and t, or raise [Not_found] if
   nothing in cache for that.  *)

let get_rrset name t = 
  try
    match cache_get name with
    | Domain values  -> let (d,rrs) = List.assoc t values in check_ttl_of d, rrs
    | NxDomain(n,time) -> check_ttl_of time, NxD n
  with 
    Expired -> ignore (purge_entry name); raise Not_found

(* Same, but look for a cname if nothing for [t] *)

let cache_lookup_cname name t =
  try 
    get_rrset name t
  with Not_found ->
    get_rrset name rrt_cname


(* Lookup for a name and type, following CNAMEs.  Return a (reverse)
   list of cnames and the final rrset as option *)

let get_follow name t =
  let rec loop acc qname =
    try
      match cache_lookup_cname qname t with
      | t,CNAME n' -> loop ((t, CNAME n')::acc) n'
      | v          -> qname, acc, Some v
    with Not_found -> qname, acc, None
  in
  loop [] name


let rec list_replace l key value =
  match l with
  | (k,v) :: rest ->
      if k=key then (key,value)::rest
      else (k,v)::list_replace rest key value
  | [] -> [(key,value)]


let add_rrset (name, ttl, rds) =
(*i  dbg_name name "add_rrset"; i*)
  let t = rrt_of_rdset_t rds
  in
  if t = rrt_nxd then
    let soa = match rds with NxD s -> s in
    Hashtbl.remove cache name;
    Hashtbl.add cache name (NxDomain(soa, of_ttl ttl))
  else
    let Domain values = (* ZZZ NxDomain ? *)
      try
	cache_get name
      with Not_found -> Domain []
    in
    Hashtbl.remove cache name;
    Hashtbl.add cache name (Domain (list_replace values (rrt_of_rdset_t rds) (of_ttl ttl, rds)))
    
let flush_cache n =
  let ks = List.filter (fun s -> is_indomain s n) (hash_keys cache) in
  let nr = List.fold_left 
      (fun acc n -> let nr = nb_rrsets (cache_get n) in Hashtbl.remove cache n; acc+nr) 0 ks;
  in
  List.length ks, nr


(* Select rrsets from cache *)

let select_all () = hash_keys cache

let select_one name =
  try ignore (cache_get name); [ name ] with Not_found -> []

let select_under name =
  let is_in d s = is_indomain s d in
  List.filter (is_in name) (hash_keys cache)

(* follow ns, cname, soa, mx, ptr *)
let select_follow name =
  let cons x xs =
    if List.mem x xs then xs else x::xs
  in
  let rec loop names acc =
      match names with
      | [] -> List.rev acc
      | n::ns -> try
	  match cache_get n with
	  | NxDomain _ -> loop ns (cons n acc)
	  | Domain v -> let more = List.flatten
		(List.map 
		   (fun (_, (_,r)) -> match r with
		   NS  nl -> nl
		   | CNAME n -> [n]
		   | SOA(n,_,_,_,_,_,_) -> [n]
		   | PTR l -> l
		   | MX  l -> List.map snd l
		   | _ -> [])
		   v)
	  in loop (ns @ more) (cons n acc)
      with Not_found -> loop ns acc
  in
  loop [name] []

(* Text representation *)

let sort_names ns = List.sort Dns.cmp_names ns

let print_cache_entry e =
  match e with
  | NxDomain(n,d) -> Printf.sprintf "%6i %s "(ttl_of d) (string_of_rdset (NxD n))
  | Domain values -> String.concat "\n\t\t\t" 
	(List.map 
	   (fun (t,(d,r)) -> 
	     let ttl = try Printf.sprintf "%6i" (ttl_of d) with Not_found -> "<expired>" in
	     Printf.sprintf "%s %s" ttl (string_of_rdset_sl r))
	   values)

let string_of_cache_entry right n e =
  Printf.sprintf (if right then format_of_string "%23s %s\n" else format_of_string "%-23s %s\n")
    (string_of_name n) (print_cache_entry e), (nb_rrsets e)

(* return text representation of cached information at name [name],
   along with the number of rrsets *)

let dump_cache_entry right name =
  try
    let e = cache_get name in
    string_of_cache_entry right name e
  with Not_found -> "", 0

let dump_select right select name =
  let names = sort_names (select name) in
  let s_n = List.map (dump_cache_entry right) (sort_names names) 
  in
  String.concat "" (List.map fst s_n),
  List.length s_n,
  List.fold_left (+) 0 (List.map snd s_n)

let dump_one    name = dump_select true select_one    name
let dump_follow name = dump_select true select_follow name
let dump_under  name = dump_select true select_under  name
let dump_all    ()   = dump_select true select_all    ()

let save_all    ()   = dump_select false select_all   ()


(*s The ns cache *)

let nscache = Hashtbl.create 5000
(* The first content of the nscache will be the root hints. *)

let find_auth n =
  let rec loop z =
    try
      let _, q = Hashtbl.find nscache z (* ZZZ ignore TTL !? *)
      in z, q
    with Not_found ->
      loop (name_parent z)
  in
  loop n
    
let add_auth (n, ttl, l) =
  Hashtbl.replace nscache n (of_ttl ttl, l)


let dump_nscache_entry right n =
  let (exp,q) = Hashtbl.find nscache n
  in
  string_of_auth right (n, ttl_of exp, q)

let dump_nscache' right =
  let names = sort_names (hash_keys nscache) in
  String.concat "\n" (List.map (dump_nscache_entry right) names), List.length names

let dump_nscache () = dump_nscache' true

let dump_nscache_all name =
  let is_in d s = is_indomain s d in
  let names = sort_names (List.filter (is_in name) (hash_keys nscache)) in
  String.concat "\n" (List.map (dump_nscache_entry true)  names), List.length names


let dump_nscache_one name =
  try 
    dump_nscache_entry true name
  with Not_found -> ""

(* duplicated from dns\_txt *)
let string_split sep s =
  let rec help s acc =
    try
      let i = String.index s sep in
      help (string_sub s (i+1)) (String.sub s 0 i :: acc)
    with Not_found ->
      List.rev (s :: acc)
  in
  help s []

let auth_of_string s =
  let nserver_of_string s =
    try
      Glue (Unix.inet_addr_of_string s)
    with Failure _ -> Extern (name_of_string s)
  in
  Scanf.sscanf s "%s %i %[^\n]" (fun n t d -> name_of_string n, t,
    List.map nserver_of_string (string_split ' ' d))

let auths_of_string s =
  let ss = string_split '\n' s in
  List.map auth_of_string ss


let nspurge () =
  List.fold_left (fun acc n -> let (d,a) = Hashtbl.find nscache n in
  acc + if ttl_of d < 0 then(Hashtbl.remove nscache n; 1) else 0)
    0
    (hash_keys nscache)
  

(*s The server cache *)

(* ip to various information
   edns unsupport and list of zones for which the server is lame *)

(* edns: time when we'll try to use edns again (0 if just now)
   [lame_for]: (expired\_time, zone) list *)

type sentry = { mutable edns:int; mutable lame_for:(int*name) list }

let dbg_ip s ip =
  dbg (Printf.sprintf "%s %s" s (Unix.string_of_inet_addr ip))

let scache = Hashtbl.create 500

let sttl = 600   (* 10 minutes *)
let lttl = 3600

let sexp_date () = int_of_float (Unix.time ()) + sttl

let lexp_date () = int_of_float (Unix.time ()) + lttl


(* The following function returns edns status and lame status for the
   given zone.  If we don't have anything in cache, we assume it
   supports edns, and is not lame!

   Since we have to check if data is stale before returning it, we
   also remove stale data from the entry, and purge the whole entry if
   all data is stale (but we don't check if lame status for other
   zones is stale).
  *)

let check_edns ent =
  if ent.edns > 0 then
    if int_of_float (Unix.time ()) < ent.edns then false
    else (ent.edns <- 0; true)
  else true

let scache_get ip zone =
  try
    let { edns=e; lame_for=l } as ent = Hashtbl.find scache ip 
    in
    let edns = check_edns ent
    in
    let islame, anylame =
      try
	let t,_ = List.find (fun (t,z) -> z=zone) l in
	if int_of_float (Unix.time ()) < t then true, true
	else begin
	  ent.lame_for <- List.filter (fun (t,z) -> z<>zone) l; 
	  false, ent.lame_for <> []
	end
      with Not_found -> false, false
    in
    if edns && not anylame then
      Hashtbl.remove scache ip;
    edns, islame
  with _ -> true, false

let scache_add_noedns ip =
  dbg_ip "S: noedns" ip;
  try
    let ent = Hashtbl.find scache ip in
    ent.edns <- (sexp_date ())
  with Not_found ->
    Hashtbl.add scache ip { edns=sexp_date(); lame_for=[] }

let add_lame ip zone =
  try
    let ent = Hashtbl.find scache ip 
    in
    ent.lame_for <- (lexp_date (),zone) :: 
      (List.filter (fun (t,z) -> z<>zone) ent.lame_for)
  with 
    Not_found ->
      Hashtbl.add scache ip { edns=0; lame_for=[lexp_date(), zone] }
      
let delay t =
  t - (int_of_float (Unix.time ()))


(* suitable for saving in a file, and reading it back later
 Displayed as a three column table:

 server ip, edns status ('-' = no info, so is believed to support,
                         integer = unsupport cached for that time)


*)
    
let dump_scache () =
  let infos = Hashtbl.fold (fun k v acc -> (k,v) :: acc) scache [] in
  let nn = List.length infos
  in
  let st = String.concat "\n" 
      (List.map
	 (fun (ip,ent) -> Printf.sprintf "%16s   %5s    %s" 
	     (Unix.string_of_inet_addr ip) 
	     (if ent.edns<>0 then (string_of_int (delay ent.edns))
	     else "-")
	     (if ent.lame_for<>[] then 
	       (String.concat " "
		  (List.map (fun (t,n) -> 
		    Printf.sprintf "%s (%i)" (string_of_name n) (delay t))
		     ent.lame_for))
	     else ""))
	 infos)
  in
  "     server          edns   lame_for\n" ^ st,
  nn
					       

(*i
let read_scache s =
  let ip,rest = Scanf.sscanf "%s %s" 
      (fun ip rest -> Unix.inet_addr_of_string ip, rest) in
  "-|%i",
  "-| zone list" -> scan_list "%s"
  i*)

(* If [check_edns] says true, there's no NoEdns info in entry. *)

let spurge_entry ip =
  let { edns=_; lame_for=l } as ent = Hashtbl.find scache ip in
  let edns = check_edns ent
  in
  let ct = int_of_float (Unix.time ())
  in
  ent.lame_for <- List.filter (fun (t,_) -> t > ct) l; 
  if edns && ent.lame_for = [] then Hashtbl.remove scache ip

let spurge () =
  List.iter spurge_entry (hash_keys scache)


(*s Saving and loading *)

let save_file file s =
  let f = open_out file in
  Printf.fprintf f "%i\n" (int_of_float (Unix.time()));
  output_string f s;
  close_out f

let save_rrsets file =
  let srr, _, nbrr = save_all () in
  save_file file srr;
  Misc.log (Printf.sprintf "Saved %i rrsets from cache." nbrr);
  nbrr

let save_auths file =
  let sns, nbns = dump_nscache' false in
  save_file file sns;
  Misc.log (Printf.sprintf "Saved %i auths from cache." nbns);
  nbns

let load_file file =
  let f = open_in file in
  let l = in_channel_length f in
  let d = int_of_string (input_line f) in
  let l = l - pos_in f in
  let b = String.create l in
  really_input f b 0 l;
  close_in f;
  int_of_float (Unix.time ()) - d, b

let load_rrsets file =
  try
    let d, s = load_file file in
    let rrsets = List.filter 
	(fun (n,t,r) -> t>0) (List.map (fun (n,t,r) -> (n,t-d,r)) (rrsets_of_string s))
    in
    let nb = List.length rrsets in
    List.iter add_rrset rrsets;
    Misc.log (Printf.sprintf "Loaded %i rrsets into cache." nb);
    nb
  with e ->
    Misc.log ("Loading rrsets failed. (" ^ Printexc.to_string e ^ ")");
    0

let load_auths file =
  try
    let d, s = load_file file in
    let auths = List.filter
	(fun (n,t,a) -> t>0) (List.map (fun (n,t,a) -> (n,t-d,a)) (auths_of_string s))
    in
    let nb = List.length auths in
    List.iter (fun (n,t,a) -> Hashtbl.replace nscache n (of_ttl t, a)) auths;
    Misc.log (Printf.sprintf "Loaded %i auth entries into cache." nb);
    nb
  with e ->
    Misc.log ("Loading auths failed. (" ^ Printexc.to_string e ^ ")");
    0
