open Dns
open Dns_txt
open Muthr
open Muthr_retry

(* tcp:connect fails *)
(* incorrect answer if rcode non 0 nor 3 ? *)
(* dig-like +[no]qr option? *)

(*s ding: a dig-like DNS querying tool *)

let print_msg m = print_string (string_of_msg m)
let error s = Printf.fprintf stderr "%s\n" s; exit 1

(* Settable parameters *)

let verb     = ref false
let port     = ref 53
let src_port = ref 0

let msg_size = ref 1024
let edns     = ref true

let rd       = ref true
let cd       = ref false
let dob      = ref false (* DO - DNSSEC OK bit *)
let nsid     = ref false

let id       = ref 0

let tcp      = ref false

let resolv_conf = "/etc/resolv.conf"

let resolvers  = ref []

let max_tries  = ref 3
let request_tm = ref 5.

(* Read conf from resolv.conf file *)

let read_conf () =
  let f = open_in resolv_conf
  in
  let rec loop acc =
    try
      let l = input_line f in
      Scanf.sscanf l "%s %s" (fun key value ->
      loop (if key="nameserver" then (Unix.inet_addr_of_string value)::acc else acc))
    with
      End_of_file -> close_in f; acc
  in
  resolvers := List.rev (loop [])

(* TODO dns.ml encode/decode allows TCP ? *)

let make_socket () =
  socket Unix.PF_INET (if !tcp then Unix.SOCK_STREAM else Unix.SOCK_DGRAM) 0

let dns_recv fd k =
  if !tcp then begin
    let b = String.create 2 in
    (* TODO muthr : really_read ? *)
    read fd b 0 2 >>= fun 2 ->
    let l = (int_of_char b.[0])*256+(int_of_char b.[1]) in
    let b = String.create l in
    read fd b 0 l >>= fun l ->
    let Unix.ADDR_INET(adr,port) = Unix.getpeername fd in
    close fd;
    let m = decode_msg b l adr port in
    k m 
  end else begin
    let b = String.create !msg_size in
    recvfrom fd b 0 !msg_size [] >>= fun (l,Unix.ADDR_INET(adr,port)) ->
    let m = decode_msg b l adr port in
    k m
  end
    
let dns_send fd m k =
  let l, b = encode_msg m in
  if !tcp then
    let b' = String.create (l+2) in
    b'.[0] <- char_of_int (l / 256);
    b'.[1] <- char_of_int (l mod 256);
    String.blit b 0 b' 2 l;
    (trywith
       (fun k -> connect fd (Unix.ADDR_INET (m.adr, m.port)) >>= fun () ->
	 really_write fd b' 0 (l+2) >>= k)
       (fun _ k  -> print_string "TCP connection failed\n";
	 sleep 100. >>= k)
       k)

  else (
    Unix.sendto fd b 0 l [] (Unix.ADDR_INET (m.adr, m.port));
    k ())


let make_query ip sname stype sclass size =

  make_query ip (if !id<>0 then !id else Random.int 64511 + 1024)
    !rd sname stype sclass
    (if !edns then Some(!dob, size, 0, if !nsid then [3,""] else []) else None)

(* wait for the answer to the question *)

let wait_answer fd q start k =
  retry_forever (fun k ->
    dns_recv fd >>= fun m ->
    if m.id = q.id && m.que = q.que && (m.rcode = 0 || m.rcode = 3) then
      k (Some (Unix.gettimeofday() -. start,m))
    else
      (print_string "< received incorrect answer\n"; print_msg m;
       k None))
    (fun e -> terminate ())
    k

(* Send a request to server and wait indefinitely for the (correct)
   answer. *)

let send_one fd qn qt server k =
  let m = make_query server qn qt 1 !msg_size
  in
  if !verb then (print_string "> sending query\n";
		 print_msg m);
  let start = Unix.gettimeofday () in
  dns_send fd m >>= fun () ->
  wait_answer fd m start >>= fun r -> k (Some r)


(* Full resolution.  Send request, wait for answer, retransmit if
   necessary. Cycle through available servers, try each one ![max_tries]
   times. *)

let do_resolv tm qn qt qc servers k =
  let fd = make_socket ()
  in
  if !src_port <> 0 then
    Unix.bind fd (Unix.ADDR_INET (Unix.inet_addr_any ,!src_port));
  retry !max_tries (fun k ->
    retry_list_tm servers tm
      (send_one fd qn qt)
      (fun () -> k None)
      (fun r -> k (Some r)))
    (fun () -> print_string "No answer received, abort.\n"; stop ())
    k
  

let ding qn qt qc =
  if !msg_size = 0 then begin
    edns := false;
    msg_size := 512
  end;
  do_resolv !request_tm (name_of_string qn) qt 1 !resolvers >>= fun (d,m) ->
  Printf.printf "\n< received DNS message (%i ms)\n" (int_of_float (d*.1000.));
  print_msg m;
    (* should check proper usage of compression? *)
  stop ()

let set_resolver sip =
  try
    let adr =
      try
	Unix.inet_addr_of_string sip
      with Failure _ ->
	(Unix.gethostbyname sip).Unix.h_addr_list.(0)
    in
    resolvers := adr :: !resolvers
  with Not_found -> error ("Can't resolve server name " ^ sip)

let qtype  = ref rrt_a
let qclass = ref 1

let spec = [("-r",   Arg.String set_resolver,
	     "server to query -- may be used multiple times.  Defaults from /etc/resolv.conf");
            ("-t",   Arg.String (fun s -> qtype := rrt_of_string s), "query type");
	    ("-m",   Arg.Set_int    max_tries,  "number of tries");
	    ("-tm",  Arg.Set_float  request_tm, "retransmission timeout");
	    ("-p",   Arg.Set_int  port, "destination port (default 53)");
	    ("-s",   Arg.Set_int  msg_size, "advertized EDNS0 size (0 no EDNS0)");
	    ("-nord", Arg.Clear rd, "do not set RD bit");
	    ("-id",  Arg.Set_int id, "query id");
	    ("-cd",  Arg.Set cd, "set CD bit");
	    ("-do",  Arg.Set dob, "set DO bit");
	    ("-tcp", Arg.Set tcp, "use TCP");
	    ("-nsid", Arg.Set nsid, "set NSID edns0 option");
	    ("-v", Arg.Set verb, "be verbose");
	    ("--src_port", Arg.Set_int src_port, "set source port")
	  ]
let usage_msg = "ding [options] name."


let main () =
  let n = ref ""
  in
  Random.self_init();
  Arg.parse spec
    (fun na -> if !n<>"" then error "only one name to query, please" else n := na)
    usage_msg;
  if !n <> "" then begin
    if !resolvers = [] then read_conf ();
    ding !n !qtype !qclass
  end
  else begin
    Arg.usage spec usage_msg;
    stop ()
  end

;;
spawn main;
start ()
