
(*s DNS names *)

(*

We define [name] as the abstract type of domain names.  The concrete
implementation is just a string in the DNS encoding (where each label
is prefixed with a length byte).

We define conversion functions to/from strings, in and sub domain
tests, getting the father domain and the root domain name.

*)

type name = string
    
let root_name = "\000"

let string_of_name n =
  let rec loop i len st =
    if i>=len then st
    else
      let ni = i + 1 + (int_of_char st.[i]) in
      st.[i] <- '.';
      loop ni len st
  in
  if n = root_name then "." 
  else
    let len = String.length n - 1 in
    loop (int_of_char n.[0]) len (String.sub n 1 len)
    
let name_of_string s =
  if s = "." or s = "" then root_name
  else
    let slen = String.length s in
    let s, len =
      if s.[slen - 1] = '.' then
	s, slen+1
      else
	s ^ ".", slen+2
    in
    let n = String.create len in
    String.blit s 0 n 1 (len-1);
  
    let rec loop i =
      try 
	let ni = String.index_from n (i+1) '.' in
	n.[i] <- char_of_int (ni - i - 1);
	loop ni
      with Not_found -> n.[i] <- '\000'
    in
    loop 0; n
    
let is_indomain s n =
  let ln = String.length n in
  let ls = String.length s in
  if ls >= ln then
    n = String.sub s (ls - ln) ln
  else false

let is_subdomain s n =
  s <> n && is_indomain s n

let name_parent n =
  let lf = int_of_char n.[0] in
  String.sub n (lf+1) (String.length n - lf - 1)

(* Simple helper: get substring of s starting at position p. *)
let string_sub s p =
  String.sub s p (String.length s - p)


(* Split a name into a list of labels, TLD first ZZZ name it better *)

let labels_of_name s =
  let rec loop s acc =
    let l = int_of_char s.[0] in
    if l=0 then acc else
    loop (string_sub s (l+1)) ((String.sub s 1 l)::acc)
  in
  loop s []

(* Compare names *)
let cmp_names n1 n2 = compare (labels_of_name n1) (labels_of_name n2)

(*s DNS resource record sets (RRsets) *)

(*

[rrt] is the abstract type of the DNS TYPE of RRsets.  It is
internally a simple [int], read by [read_rrt] defined in the decode
section below.

*)

type rrt = int

let rrt_a     = 1
let rrt_ns    = 2
let rrt_cname = 5
let rrt_soa   = 6
let rrt_ptr   = 12
let rrt_mx    = 15
let rrt_txt   = 16
let rrt_aaaa  = 28
let rrt_opt   = 41
let rrt_any   = 255 (* valid only for qtype *)
let rrt_no    = -1
let rrt_nxd   = -2

let string_types =  [
  ("A", rrt_a); ("NS", rrt_ns); ("CNAME", rrt_cname); ("SOA", rrt_soa);
  ("PTR", rrt_ptr); ("MX", rrt_mx); ("TXT", rrt_txt); ("AAAA", rrt_aaaa);
  ("ANY", rrt_any); ("!", rrt_no); ("NXD", rrt_nxd)
]

let rrt_of_string st =
  try
    List.assoc (String.uppercase st) string_types
  with Not_found -> 
    try
      Scanf.sscanf st "TYPE%i" (fun i -> i)
    with _ -> failwith ("unknown type " ^ st)

let string_of_rrt t =
  try 
    List.assoc t (List.map (fun (a,b) -> b,a) string_types)
  with Not_found -> Printf.sprintf "TYPE%i" t

(* [rdset_t] is the type of rdatas from an RRset, including the type.
   It's a sum type with the more common types and the [TypeOther] case
   for all other types.  *)

type rdset_t   = 
    A     of Unix.inet_addr list
  | NS    of name list
  | CNAME of name
  | SOA   of name * name * int * int * int * int * int
  | PTR   of name list
  | MX    of (int * name) list
  | TXT   of string list list
  | AAAA  of Unix.inet_addr list
  | OPT   of string
  | TypeOther of int * string list
  | Bad   of rrt * string list

  | No    of rrt * name
  | NxD   of name

let rrt_of_rdset_t rr =
  match rr with
  | A     _ -> rrt_a
  | NS    _ -> rrt_ns
  | CNAME _ -> rrt_cname
  | SOA   _ -> rrt_soa
  | PTR   _ -> rrt_ptr
  | MX    _ -> rrt_mx
  | TXT   _ -> rrt_txt
  | AAAA  _ -> rrt_aaaa
  | OPT   _ -> rrt_opt
  | TypeOther (i,_) -> ((Obj.magic i):rrt)
  | No(t,_) -> t (* ZZZ rrt no ? *)
  | NxD _   -> rrt_nxd (*failwith "rrt of rdset t NxD"*)

(* parametric polymorphism gets in our way here! *)

let count_rr rrsl =
  let count rrs =
    match rrs with
    | A l -> List.length l
    | NS l | PTR l -> List.length l
    | MX l -> List.length l
    | TXT l -> List.length l
    | AAAA l -> List.length l
    | TypeOther(_,l) -> List.length l
    | _ -> 1
  in
  List.fold_left (fun c (_,_,rrs) -> c + count rrs) 0 rrsl

(* An rrset is made of an owner name, a TTL, and a rdset. *)
type rrset_t = name * int * rdset_t

type class_t = IN | Class_other of int


(*s DNS messages *)

(* 

  This is the type of DNS messages.  It is either produced from wire
  encoding by the [decode_msg] function or built in order to be ZZZ.

  [adr] and [port] are the source if the [msg] has been received then
  built by [decode_msg], the destination if it is to be sent.

  [len] is meaningful only on a received message (actual length of a
  to be sent message will be computed when encoding it).

  *)

type parse_error = 
  | BadRdata of int
  | BadClass of int
  | BadFormat of string


type msg = {
    adr:Unix.inet_addr;
    port:int;
    len:int;
    bad:(parse_error * string) option;
    id:int;
    opcode:int;
    qr:bool; aa:bool; tc:bool; rd:bool; ra:bool; ad:bool; cd:bool;
    rcode:int;
    (* EDNS do, payload size, rcodeext, (attribute,value) list *)
    edns0:(bool * int * int * (int*string) list) option;
    qdcount:int; ancount:int; nscount:int; arcount:int;
    que:(name * int * int) list;            (* type, class *)
    ans: rrset_t list;
    aut: rrset_t list;
    add: rrset_t list;
  }

let make_query ip id rd name qtype qclass edns0 =
  { adr=ip; port=53; len=0; bad=None; id=id; opcode=0; qr=false; aa=false; tc=false;
    rd=rd; ra=false; ad=false; cd=false; rcode=0;
    edns0 = edns0;
    qdcount=1; ancount=0; nscount=0; arcount=0;
    que=[ (name, qtype, qclass) ];
    ans = []; aut = []; add = [];
  }

(*
  Make an answer based on the received request.  The request provides
  address/port and query to be used in the answer as well as EDNS support
*)

let make_empty_answer m rcode ms =
  let edns = match m.edns0 with
  | None -> None
  | Some _ -> Some(false, ms, 0, [])
  in
  { adr=m.adr; port=m.port; len=m.len; bad=m.bad; id=m.id; opcode=0; qr=true; aa=false;
    tc=false; rd=m.rd; ra=true; ad=false; cd=false; rcode=rcode; edns0=edns;
    qdcount=1; ancount=0; nscount=0; arcount=0;
    que=m.que; ans=[]; aut=[]; add=[]
  }

let make_answer aa ra m rcode ms ans aut add nsid =
  let edns = match m.edns0 with
  | None -> None
  | Some (_,ms, 0, [3,""]) -> Some(false, ms, 0, [3, nsid])
  | Some _ -> Some(false, ms, 0, [])
  in
  { adr=m.adr; port=m.port; len=0; bad=m.bad; id=m.id; opcode=0; qr=true;
    aa=aa; tc=false; rd=m.rd; ra=ra; ad=false; cd=false;
    rcode=rcode; edns0=edns; qdcount=1; ancount=count_rr ans;
    nscount=count_rr aut; arcount=count_rr add; que=m.que; ans=ans;
    aut=aut; add=add }

let make_aut_answer m rcode ms ans aut add =
  make_answer true false m rcode ms ans aut add

let make_res_answer m rcode ms ans aut add =
  make_answer false true m rcode ms ans aut add




(* To perform compression, we'll need to turn a [name] into a list of
   labels, in reverse order (TLD first).  [list_of_name] and
   [name_of_list] allow conversion between a [name] and a list of
   labels.  *)


let list_of_name s =
  let rec loop s acc =
    let l = int_of_char s.[0] in
    if l=0 then acc else
    loop (string_sub s (l+1)) ((String.sub s 0 (l+1) )::acc)
  in
  loop s []

let name_of_list l =
  let rec loop l acc =
    match l with
    | [] -> acc
    | h::t -> loop t (h ^ acc)
  in
  loop l ""


(*s Decoding a message *)

(*

  We define various [get_] and [read_] functions.  [get_] extract an
  information at a given offset from the message buffer.  [read_]
  are given a \emph{position} (which is of type [int ref]), they
  extract the information \textbf{and} move the position to the next
  element.

  We will perform some syntactic checks on the rdatas, if one fails
  the [BadRdata] exception will be raised, with the offending offset
  as parameter.

  We'll also refuse to parse messages where RRsets for classes other
  then IN appear, by raising the [BadClass] exception.

  ZZZ String.bad offset ZZZ

*)

(* TODO: we should add a lot of sanity checks on the received
   messages. *)

exception ParseError of parse_error

(* Get 16 bit word coded in buffer string [b] at position [i]. *)
let get_16 b i =
  int_of_char b.[i] lsl 8 + int_of_char b.[i+1]

let get_32 b i =
  let t1 = get_16 b i in
  let t2 = get_16 b (i+2) in
  (t1 lsl 16) + t2

let get_data b (o,l) =
  String.sub b o l

(* read a 16 bit word *)

let read_16 b p =
  let i = !p in
  incr p; incr p;
  get_16 b i

(* read a 32 bit word *)
let read_32 b p =
  let i = !p in
  p := !p + 4;
  get_32 b i

(* data is prefixed by its length on 16 bits -- ZZZ unused *)
let read_data b p =
  let len = read_16 b p in
  let s = String.sub b !p len
  in p := !p + len; s

let is_pointer b o =
  int_of_char b.[o] > 63


(*
   [read_name] extract as a [name] the name starting at position [p]
   in buffer [b], taking care of decompression (ie following
   pointers). Updates [p].

*)

let get_label b o =
  let l = int_of_char b.[o] in
  String.sub b o (l+1)

let dns_lower c =
  if (c >= 'A') && (c <= 'Z') then
      char_of_int (int_of_char c + 32)
  else c

let get_label b o =
  let l = int_of_char b.[o] in
  let s = String.sub b o (l+1) in
  for i=1 to l do
    s.[i] <- dns_lower s.[i]
  done;
  s
  

let get_pointer b o n1 =
  let n1 = n1 land 63 in
  let n2 = int_of_char b.[o+1] 
  in
  (n1 lsl 8) + n2

let read_name b p =
  let rec loop o acc =
    let l = int_of_char b.[o] 
    in
    if l = 0 then ("\000"::acc), o+1 else
    if l<=63 then 
      let label = get_label b o in
      loop (o+l+1) (label::acc)
    else
      let res, _ = loop (get_pointer b o l) acc in res, o+2
  in
  let res, newp = loop !p [] in
  p := newp; (Obj.magic (name_of_list res):name)



(* Return offset of [name] at position [p], that is, offset of actual
   string, which is either [p] or the pointed offset. Update [p].

   As a special case, offset of the root domain is returned as 0.

   [follow_pointer] is recursive to handle the (unlikely) case where a
   pointer points on another pointer.  We want to reach the actual
   label string in the message. ZZZ forbidden ?

   [name_length] computes recursively the space taken by the name at
   the considered position.  It's 1 for an empty label, 2 for a
   pointer, or the size of the label plus the length of what follows..

   *)

let read_name_offset b p =
  let rec follow_pointer o =
    let l = int_of_char b.[o] in
    if l=0 then 0 else
    if l<=63 then o
    else follow_pointer ((l land 63) lsl 8 + int_of_char b.[o+1])
  in
  let rec name_length o =
    if b.[o] = '\000' then o+1 else
    if is_pointer b o then o+2
    else name_length (o+1+ int_of_char b.[o])

  in  
  let off = follow_pointer !p in
  p := name_length !p;
  off


(*s Reading rrsets. *)

(*
   The main work is extracting rdsets.

   All rdata is prefixed by its rdlen.  However, some types have "self
   delimited rdatas", and we should check the advertised rdlen is
   indeed correct.

   [check current_offset initial_offset len data] does this.  It
   simply returns [data] if all is ok, otherwise it raises a
   [BadRdata] exception.

   Some RR types can only have singleton rrsets, we'll check this with
   [check_singleton].

   [get_name] is for extracting a [name] as a component of the rdata
   (rdlen then considers the whole rdata, not the [name] itself),
   while [get_dom] extracts a [name] as the single rdata component and
   so checks the rdlen.

   We define a set of functions extracting an rdata field from the
   message, for the main RR types, all checking the rdlen is correct.

   For RR types whose rdata format we don't know, we obviously can't
   check this.

*)


let check co io len data =
  if co = io+len then data else raise (ParseError (BadRdata io))

let check_singleton ols =
  match ols with
  | [ o,l ] -> o,l
  | _ -> failwith "not a singleton list"

let get_name b o =
  if o=0 then root_name else read_name b (ref o)

let get_dom  b (o,rdl) = 
  if o=0 then root_name else 
  let p = ref o in 
  let n = read_name b p
  in check !p o rdl n

let get_a    b (o,rdl) : Unix.inet_addr = 
  if rdl = 4 then Obj.magic (String.sub b o 4)
  else raise (ParseError (BadRdata o))

let get_aaaa b (o,rdl) : Unix.inet_addr = 
  if rdl = 16 then Obj.magic (String.sub b o 16)
  else raise (ParseError (BadRdata o))

let get_soa  b (o,rdl) = 
  let p = ref o in
  let mname   = read_name b p in
  let rname   = read_name b p in
  let serial  = read_32 b p in
  let refresh = read_32 b p in
  let retry   = read_32 b p in
  let expire  = read_32 b p in
  let min     = read_32 b p
  in
  check !p o rdl (SOA(mname, rname, serial, refresh, retry, expire, min))

let get_mx b (o,rdl) =
  let p = ref o in
  let pref = read_16 b p in
  let exch = read_name b p in 
  check !p o rdl (pref, exch)


(* [get_txt] is a bit more complex since the rdata of a TXT rr
   actually is a \emph{list} of strings. *)

let get_txt b (o,rdl) =
  let rec loop co acc =
    if co > o+rdl then raise (ParseError (BadRdata o))
    else if co = o + rdl then acc else
    let len = int_of_char b.[co] in
    loop (co+len+1) ((String.sub b (co+1) len)::acc)
  in
  List.rev (loop o [])

(*

 [get_rdset] uses the previous functions to build a rdset from the
  RR type and the list of (offset, len) pairs describing the rdatas.

*)
let get_rdset b t ols =
    match Obj.magic t with
       1 -> A  (List.map (get_a b) ols)
    |  2 -> NS (List.map (get_dom b) ols)
    |  5 -> CNAME (get_dom b (check_singleton ols))
    |  6 -> get_soa b (check_singleton ols)
    | 12 -> PTR (List.map (get_dom b) ols)
    | 15 -> MX (List.map (get_mx b) ols)
    | 16 -> TXT (List.map (get_txt b) ols)
    | 28 -> AAAA (List.map (get_aaaa b) ols)
    | 41 -> OPT (get_data b (check_singleton ols))
    |  i -> TypeOther(i, List.map (get_data b) ols)

(*i
(* 

   We apparently have a set of [n] rrs in the message at position [p].
   Read them as a list of rrsets.

   An OPT pseudo RR can only occur in the additional section and
   encodes an extension of the message header.

   [accept_opt] has type 'a option ref option. If None, an OPT RR
   would be illegal here.  Otherwise, Some r with r a ZZZ/

   The main work here is to group RRs that have the same name and type
   in a single RRset.  It seems to be a design error that DNS messages
   carry RRs and not RRsets.  However, since name compression for
   owner names is mandatory (as specified in rfc1123-6.1.2.4, was only
   optional in rfc1035-4.1.4) we can identify names from their offset
   (from [read_name_offset] defined above) and assemble RRs from a
   single RRset by comparing name offset and rrt.

   Some servers don't use name compression, RRsets will not be
   reassembled correctly in that case.  This can be detected and
   corrected later in the message processing flow.

   There's one special case where compression is not used, the root
   domain (indeed it would not be clever to replace single byte null
   labels by two-byte pointers in order to compress).
   [read_name_offset] gave it special offset 0 by convention (that
   can't be a valid offset since it points at the beginning of the
   header).

   An [acc_t] value is an RRset in construction. [nof] is the name
   offset, [t] the type, [data] the list of (data offset, rdlen)
   pairs.

   rfc2181 (5.2) states it is an error for all RRs in a RRset not to
   have the same ttl.  However, if coming from an authoritative
   source, we should align on the lowest of the ttls.  We choose to do
   that in any case.

   TODO: should elaborate on authority (glue ?)

   *)


type acc_t = 
    { nof:int; t:rrt; mutable ttl:int; mutable data:(int*int) list }

let bobit w n = (w land (1 lsl n)) <> 0
let setbit w n = w := !w lor (1 lsl n)

let read_rrt b p =
  (Obj.magic read_16 b p : rrt)

let read_rrsets b p n accept_opt =

  let handle_opt (nof, c, ttl, dof, rdl) =
    match accept_opt with
    | None -> failwith "OPT in unauthorized section"
    | Some edns0 ->
	if nof <> 0 then failwith "incorrect name for OPT";
	if (ttl land 0xff) = 0 then
	  edns0 := Some(bobit ttl 15, c, (ttl lsr 24), get_data b (dof,rdl))
  in
  
  (* [acc] is an [acc_t] list *)
  let rec loop n acc p =
    if n = 0 then acc else
    let nof = read_name_offset b p in
    let t   = read_rrt  b p in
    let c   = read_16 b p in
    let ttl = read_32   b p in
    let rdl = read_16   b p in
    let dof = !p in
    if t = rrt_opt then begin
      handle_opt (nof, c, ttl, dof, rdl);
      loop (n-1) acc p
    end
    else if c <> 1 then raise (BadClass t)
    else
    try 
      let rrs = List.find (fun r -> r.nof = nof && r.t = t) acc
      in 
      if rrs.ttl > ttl then rrs.ttl <- ttl;
      rrs.data <- (dof,rdl) :: rrs.data;
      p := !p + rdl;
      loop (n-1) acc p
    with
      Not_found ->
	p := !p + rdl;
	loop (n-1) ({ nof=nof; t=t; ttl=ttl; data=[dof,rdl] }::acc) p
  in

(* 

   Once we have the list of [acc_t] we turn it into a list of RRsets.
   Basically, extract name and data from the corresponding offsets.
   One fine point about classes: rfc1035-3.3 says the CNAME, MX, NS,
   PTR, SOA and TXT RRs (plus others that don't appear explicitly in
   our [rrs_t] type) are expected to appear in all classes. [A] (and
   [AAAA] ZZZ) should not, but what if we actually receive a msg with
   an [A] type for another class?  But, hey, we just raised an
   exception when we saw a non IN class, so we don't need to care
   about this lousy stuff. Good.

*)

  List.rev (List.map 
	      (fun rrs ->
		let name = get_name b rrs.nof in
		(name, rrs.ttl, get_rrdatas b rrs.t rrs.data))
	      (loop n [] p))

i*)
(* new version *)



(* 

   We apparently have a set of [n] RRs in the message at position [p].
   Read them as a list of rrsets.

   An OPT pseudo RR can only occur in the additional section and
   encodes an extension of the message header.

   [accept_opt] has type 'a option ref option. If None, an OPT RR
   would be illegal here.  Otherwise, Some r with r a ZZZ/

   The main work here is to group RRs that have the same name and type
   in a single RRset.  It seems to be a design error that DNS messages
   carry RRs and not RRsets.  However, since name compression for
   owner names is mandatory (as specified in rfc1123-6.1.2.4, was only
   optional in rfc1035-4.1.4) we can identify names from their offset
   (from [read_name_offset] defined above) and assemble RRs from a
   single RRset by comparing name offset and rrt.

   This sounds good except it's not.  Some servers don't perform name
   compression contrary to rfc1123-6.1.2.4 requirement.

   Thus if we find no matching offset we'll need to compare names.  We
   don't want to extract names if it's not necessary.  It's necessary
   only when the name is not fully compressed. ZZZ false assumption,
   see Trace/comp.txt|cap

   [name_offset_t] contains either an offset if the name is coded as a
   pointer or its position and value if it's not.

   *)


type name_offset_t = 
  | Pointer of int
  | Name    of int * name

(* Return offset of [name] at position [p], that is, offset of actual
   string, which is either [p] or the pointed offset. Update [p].

   As a special case, offset of the root domain is returned as 0.

   [follow_pointer] is recursive to handle the (unlikely) case where a
   pointer points on another pointer.  We want to reach the actual
   label string in the message. ZZZ forbidden ?

   [name_length] computes recursively the space taken by the name at
   the considered position.  It's 1 for an empty label, 2 for a
   pointer, or the size of the label plus the length of what follows..

   *)

let read_name_offset b p =
  let rec follow_pointer o =
    let l = int_of_char b.[o] in
    if l=0 then Pointer 0 else
    if l<=63 then Name(o, get_name b o)
    else let p = (l land 63) lsl 8 + int_of_char b.[o+1]
    in if is_pointer b p then follow_pointer p else Pointer p
  in
  let rec name_length o =
    if b.[o] = '\000' then o+1 else
    if is_pointer b o then o+2
    else name_length (o+1+ int_of_char b.[o])

  in  
  let off = follow_pointer !p in
  p := name_length !p;
  off


(*

   There's one special case where compression is not used, the root
   domain (indeed it would not be clever to replace single byte null
   labels by two-byte pointers in order to compress).
   [read_name_offset] gave it special offset 0 by convention (that
   can't be a valid offset since it points at the beginning of the
   header).

   An [acc_t] value is an RRset in construction. [nof] is the name
   offset, [t] the type, [data] the list of (data offset, rdlen)
   pairs.

   TODO: should elaborate on authority (glue ?)
*)

type acc_t = 
    { nof:int; nam:name; t:rrt; mutable ttl:int; mutable data:(int*int) list }

let bobit  w n = (w land (1 lsl n)) <> 0
let setbit w n = w := !w lor (1 lsl n)

let read_rrt b p =
  (Obj.magic read_16 b p : rrt)

let read_edns_options b p rdl =
  let limit = !p + rdl
  in
  let rec loop acc =
    if !p >= limit then acc else
    let code = read_16 b p in
    let v    = read_data b p in
    loop ((code,v)::acc)
  in
  loop []

let read_rrsets b p n accept_opt =

  let handle_opt (nof, c, ttl, dof, rdl) =
    match accept_opt with
    | None -> failwith "OPT in unauthorized section"
    | Some edns0 ->
	if nof <> (Pointer 0) then failwith "incorrect name for OPT";
	if (ttl land 0xff) = 0 then
	  edns0 := Some(bobit ttl 15, c, (ttl lsr 24), read_edns_options b p rdl)
  in
  
  (* [acc] is an [acc_t] list *)
  let rec loop n acc p =
    if n = 0 then acc else
    let nof = read_name_offset b p in       (* read one RR *)
    let t   = read_rrt  b p in
    let c   = read_16 b p in
    let ttl = read_32   b p in
    let rdl = read_16   b p in
    let dof = !p in

    if t = rrt_opt then begin               (* special OPT case *)
      handle_opt (nof, c, ttl, dof, rdl);
      loop (n-1) acc p
    end
    else if c <> 1 then raise (ParseError(BadClass t))  (* bad class *)
    else begin

    p := !p + rdl;                          (* normal case *)
    match nof with
    | Pointer nof ->
	(* This name is compressed *)
	(try 
	  (* Does this RR belong to one of the started rrsets? *)
	  let rrs = List.find (fun r -> r.nof = nof && r.t = t) acc
	  in 

          (*
	    rfc2181-5.2 states it is an error for all RRs in a RRset not to
	    have the same ttl, but we should accept it if coming from an
	    authoritative source, and align on the lowest of the ttls.
	    
	    We choose to do that in any case. ZZZ check aa?   *)

	  if rrs.ttl > ttl then rrs.ttl <- ttl;
	  rrs.data <- (dof,rdl) :: rrs.data;
	  loop (n-1) acc p
	with
	  Not_found ->
	    (* No, add a new rrset to our list *)
	    let nam = get_name b nof in
	    loop (n-1) ({ nof=nof; nam=nam; t=t; ttl=ttl; data=[dof,rdl] }::acc) p
	)
    | Name(nof,nam) ->
	(* This one is not.  It may still belong to an rrset. *)
	try
	  let rrs = List.find (fun r -> r.nam = nam & r.t = t) acc
	  in
	  (* I'd just let you fall on the floor, you msg from stupid server *)
	  if rrs.ttl > ttl then rrs.ttl <- ttl;
	  rrs.data <- (dof,rdl) :: rrs.data;
	  loop (n-1) acc p
	with
	  Not_found ->
	    loop (n-1) ({ nof=nof; nam=nam; t=t; ttl=ttl; data=[dof,rdl] }::acc) p
    end
  in

(* 

   Once we have the list of [acc_t] we turn it into a list of RRsets.
   Basically, extract data from the corresponding offsets (we've
   already extracted names).  One fine point about classes:
   rfc1035-3.3 says the CNAME, MX, NS, PTR, SOA and TXT RRs (plus
   others that don't appear explicitly in our [rrs_t] type) are
   expected to appear in all classes. [A] (and [AAAA] ZZZ) should not,
   but what if we actually receive a msg with an [A] type for another
   class?  But, hey, we just raised an exception when we saw a non IN
   class, so we don't need to care about this lousy stuff.  Sounds
   good.

*)

  List.rev (List.map 
	      (fun rrs ->
		(rrs.nam, rrs.ttl, get_rdset b rrs.t rrs.data))
	      (loop n [] p))


(*s Read questions, decode message *)

(* Read the questions.  Nobody seems to currently support multiple
   questions, and no semantics are defined for it, but it is
   nevertheless syntactically correct.  We're good boys so we won't
   refuse to parse multiple questions messages even if we'll
   `probably' reject them later in the processing flow.  *)

let read_que b p count =
  let rec loop acc c =
    if c = 0 then List.rev acc
    else
      let qname = read_name b p in
      let qtype = read_16 b p in
      let qclass = read_16 b p 
      in
      if qclass <> 1 then raise (ParseError (BadClass !p)) else
      loop ((qname, qtype, qclass)::acc) (c-1)
  in
  loop [] count

let err_msg adr port l e b =
  { adr=adr; port=port; len=l; bad=Some(e,b); id=0; opcode=0;
    qr=false; aa=false; tc=false; rd=false; ra=false;
    ad=false; cd=false; rcode=0; edns0=None;
    qdcount=0; ancount=0; nscount=0; arcount=0;
    que=[]; ans=[]; aut=[]; add=[]
  }

let decode_msg b l adr port =
  try
    let p = ref 0 in
    let id = read_16 b p in
    let w2 = read_16 b p in
    let bit n = (w2 land (1 lsl n)) lsr n (* n reversed from rfc1034 header *)
    in
    let qr = bit 15 in
    let op = (w2 lsr 11) land 15 in
    let aa = bit 10 in
    let tc = bit 9 in
    let rd = bit 8 in
    let ra = bit 7 in
    let ad = bit 5 in
    let cd = bit 4 in
(*    let qr = (w2 land (1 lsl 15)) lsr 15 in
    let op = (w2 lsr 11) land 15 in
    let aa = (w2 land (1 lsl 10)) lsr 10 in
    let tc = (w2 land (1 lsl 9)) lsr 9 in
    let rd = (w2 land (1 lsl 8)) lsr 8 in
    let ra = (w2 land (1 lsl 7)) lsr 7 in
    let ad = (w2 land (1 lsl 5)) lsr 5 in
    let cd = (w2 land (1 lsl 4)) lsr 4 in*)
    let rcode = (w2 land 15) in
    let qdcount = read_16 b p in
    let ancount = read_16 b p in
    let nscount = read_16 b p in
    let arcount = read_16 b p in
    let edns0 = ref None
    in
    let que = read_que    b p qdcount      in
    let ans = read_rrsets b p ancount None in
    let aut = read_rrsets b p nscount None in 
    let add = read_rrsets b p arcount (Some edns0) in

    (* ZZZ check !p = l ? *)
    
    { adr=adr; port=port; len=l; bad=None; id=id; opcode=op;
      qr=(qr<>0); aa=(aa<>0); tc=(tc<>0); rd=(rd<>0); ra=(ra<>0); 
      ad=(ad<>0); cd=(cd<>0); rcode=rcode; edns0= !edns0;
      qdcount=qdcount; ancount=ancount; nscount=nscount; 
      arcount= if !edns0 = None then arcount else arcount-1;
      que = que; ans = ans; aut = aut; add = add;
    }
  with 
    ParseError e -> err_msg adr port l e b
  | e -> err_msg adr port l (BadFormat (Printexc.to_string e)) b


(*s DNS compression *)

(* Some [name]s must be compressed in messages.  The labels are
   prefixed by a length octet or replaced by a pointer.  We define
   three functions taking a message buffer and position as an [int
   ref].  They all write in the buffer at the specified position and
   update the position.

   [write_label] writes a straight label, prefixed by its length
   (maximum 63),

   [write_pointer] writes a pointer to the given offset.  A pointer is
   coded on two bytes, with the two msbs of the first byte set to 1,

   [write_null] writes a null label, it's a straight label with 0
   characters, so made of a single 0 byte.

  *)

let write_label n b p =
  let l = int_of_char n.[0] in
  String.blit n 0 b !p (l+1);
  p := !p + l + 1

let write_pointer off b p =
  let n1 = 192 lor (off / 256)
  and n2 = off mod 256
  in
  b.[!p] <- char_of_int n1; incr p;
  b.[!p] <- char_of_int n2; incr p

let write_null b p =
  b.[!p] <- '\000'; incr p


(* Compression is performed while writing names in the message buffer.
   A compression tree is associated with the buffer.  Each node in
   tree has an offset field giving the position of the corresponding
   label in the buffer, and a list of son nodes.  The root node has 0
   as offset by convention (no real offset can be 0 since that points
   in the header of the message.)  
 *)

type node = 
    { mutable offset : int;
      mutable sons : node list }



(* Check wether label l is equal to label at offset [o] in b. *)

let eq_label lb b o =
  let l = int_of_char b.[o] in
  lb = String.sub b o (l+1)

(* This function is for debugging.  It prints the compression tree
   (given as its root node [t] and the buffer [b]), printing for each
   node the label occuring at the node offset.  Uses indentation to
   convey the hierarchical relation of nodes.
*)

let print_tree t b =
  let rec loop t i =
    match t with
      { offset = o; sons = s } ->
	Printf.printf "%s%s (%i)\n" (String.make i ' ') (get_label b o) o;
	List.iter (fun t -> loop t (i+5)) s
  in
  List.iter (fun t -> loop t 0) t.sons

(* This is the function to add a name in the message buffer,
   compressing it if possible.  [buf] is the message buffer, [p] a
   [ref] to the current position, [t] the root of the compression
   tree.

   When done, the name has been added at position [p] (compressed if
   possible), [p] points at the position following the name, and [t]
   has been updated.

   We first try to follow an existing branch matching a suffix of name
   ([follow_branch]).  If we process name completely this way, we just
   write a pointer.  Otherwise we continue by adding a new branch
   ([add_branch]) then write a pointer to the suffix we just matched.
   If we matched nothing, we add a whole branch and terminate it with
   a null label.

*)

let write_name t name b p =

  (* Add a new branch: since the list is reversed, we first recurse
     until at the first label, then we build nodes and write labels in
     [b] when returing from recursion, processing them in the
     ``natural'' order.

     The root name is a special case: we always encode it as a null
     pointer.
   *)

  let rec add_branch name =

    match name with
    | [ n ] ->
	let t = { offset = !p; sons = [] } in
	write_label n b p; t
    | n::ns ->
	let t = add_branch ns in
	let t = { offset = !p; sons = [t] } in
	write_label n b p; t
  in

  (* Follow branch for compressing name. *)
  let rec follow_branch t name =
    match name with
    | [] -> write_pointer t.offset b p
    | n::ns ->
	try
	  let t = List.find (fun s -> eq_label n b s.offset) t.sons
	  in follow_branch t ns
	with
	  Not_found ->
	    let br = add_branch name in
	    t.sons <- br :: t.sons;
	    if t.offset <> 0 then
	      write_pointer t.offset b p
	    else write_null b p (* we followed nothing *)
  in
  if name=root_name then write_null b p else
  follow_branch t (list_of_name name)
;;




(*s Building a message *)

(*

  Now that we can write compressed [name]s in the message, we define
  functions to write other data and use them for [encode_msg].

  rfc1123-6.1.3.5 specifies that [name]s must be compressed for
  "well-known" RR types and must not for others (so that new unknown
  RRs can be correctly processed by caches).  rfc3597-4 precises that
  these well-known RR types are those defined in rfc1035.  Only such
  well-known RRs types have their own entry in our [rrs_t] type and
  will use [write_name] when appropriate. Other are represented as
  [TypeOther] and have their rdata encoded through [write_other].
  This ensures we don't compress [name]s we shouldn't.

 *)


let put_16 b o n =
  b.[o]   <- char_of_int ((n land 0xFF00) lsr 8);
  b.[o+1] <- char_of_int  (n land 0x00FF)

let write_16 b p n =
  b.[!p] <- char_of_int ((n land 0xFF00) lsr 8); incr p;
  b.[!p] <- char_of_int  (n land 0x00FF); incr p
  
let write_32 b p n =
  write_16 b p ((n land 0xFFFF0000) lsr 16);
  write_16 b p  (n land 0x0000FFFF)

let start_msg l =
  let buf = String.make l '\000'
  and root = { offset = 0; sons = [] }
  and p = ref 0
  in (buf, p, root)

(* Makes a query with specified id and question.  [edns0] is the
   advertised msg size, if any.  Only for non recursive queries. *)

let make_a_query id rd name qtype qclass edns0 =
  let l = String.length name in
  let (b, p, root) = start_msg (l+16+(if edns0<>0 then 11 else 0))
  in
  put_16 b 0 id;
  b.[2] <- if rd then '\001' else '\000';
  b.[5] <- '\001';
  p := 12;
  String.blit name 0 b !p l; p := !p + l;
  write_16 b p qtype;
  write_16 b p qclass;
  if edns0<>0 then begin
    write_null b p;
    write_16 b p 41;
    write_16 b p edns0;
    b.[11] <- '\001'   (* One -- pseudo -- RR in Additionnal section *)
  end;
  b


(* Encoding a message *)

let write_a b p ip =
  write_16 b p 4;
  String.blit (Obj.magic ip) 0 b !p 4;
  p := !p + 4

let write_aaaa b p ip =
  write_16 b p 16;
  String.blit (Obj.magic ip) 0 b !p 16;
  p := !p + 16

(* We write rdlen after the name itself since its value will be
   determined by name compression. *)

let write_rdname root b p n =
  let bo = !p in p := !p + 2;
  write_name root n b p;
  put_16 b bo (!p-bo-2)


let write_txt b p ts =
  let o = !p in
  let rec loop tl ts =
    match ts with
    | [] -> tl
    | h::ts -> let l = String.length h in
      b.[!p] <- char_of_int l; incr p;
      String.blit h 0 b !p l;
      p := !p + l;
      loop (tl+l+1) ts
  in
  p := !p + 2;
  let tl = loop 0 ts in
  put_16 b o tl

let write_mx root b p (pref,exch) =
  let bo = !p in p := !p + 2;
  write_16 b p pref;
  write_name root exch b p;
  put_16 b bo (!p-bo-2)

let write_other b p n =
  let l = String.length n in
  write_16 b p l;
  String.blit n 0 b !p l; p := !p+l

let write_rrs b p t c ttl rds nof f =
  List.iter (fun rd ->
    write_pointer nof b p; write_16 b p t; write_16 b p c; write_32 b p ttl;
    f b p rd)
    rds

(* Write the part of an rrset starting at the first rdlen, ie, rdlen,
   rdata, then a pointer to the first name ([nof]), copy of the type,
   class and ttl, rdlen and rdata for all 'other' RRs in the rrset.
 *)

let write_rrd root b p nof t c ttl rrd =
  let write_rdname = write_rdname root in
  let wr f b p n t c ttl rds nof =
    f b p n; write_rrs b p t c ttl rds nof f
  in
  match rrd with
  | A (ip::ips)  -> wr write_a b p ip t c ttl ips nof
  | NS (ns::nss) -> wr write_rdname b p ns t c ttl nss nof
  | CNAME n'     -> write_rdname b p n'

  | SOA(m,r,s,ref,ret,exp,min)  ->
      let bo = !p in p := !p + 2;
      write_name root m b p; write_name root r b p;
      write_32 b p s; write_32 b p ref;
      write_32 b p ret; write_32 b p exp;
      write_32 b p min;
      put_16 b bo (!p-bo-2)

  | PTR (n::ns)    -> wr write_rdname b p n t c ttl ns nof
  | MX (m::ms)     -> wr (write_mx root) b p m t c ttl ms nof 
  | TXT (tx::txs)  -> wr write_txt b p tx t c ttl txs nof
  | AAAA (ip::ips) -> wr write_aaaa b p ip t c ttl ips nof
  | OPT _ -> failwith "OPT RR in write_rrd"

  | TypeOther(i,s::ss) -> wr write_other b p s t c ttl ss nof


let write_options b p options =
  let o = !p in p := !p + 2;
  let rec loop ls =
    match ls with
    | [] -> ()
    | (c,v)::ls ->
	write_16 b p c;
	write_other b p v;
	loop ls
  in
  loop options;
  put_16 b o (!p-o-2)

let encode_msg m =
  let (b, p, root) = start_msg 1024 (* ZZZ m.len*)
  in
  write_16 b p m.id;
  let w2 = ref 0 in
  if m.qr then w2 := !w2 lor (1 lsl 15);
  if m.aa then w2 := !w2 lor (1 lsl 10);
  if m.tc then w2 := !w2 lor (1 lsl 9);
  if m.rd then w2 := !w2 lor (1 lsl 8);
  if m.ra then w2 := !w2 lor (1 lsl 7);
  if m.ad then w2 := !w2 lor (1 lsl 5);
  if m.cd then w2 := !w2 lor (1 lsl 4);

  let opt = if m.edns0 = None then 0 else 1
  in
  List.iter (write_16 b p) [ (!w2 lor (m.rcode land 15));
			     m.qdcount; m.ancount; m.nscount; m.arcount+opt ];
  (* Question section *)
  List.iter (fun (qn, qt, qc) ->
    write_name root qn b p;
    write_16 b p qt; write_16 b p qc)
    m.que;
  (* Other sections *)
  List.iter
    (fun (n,ttl,rrd) ->
      let nof = !p in
      let t = rrt_of_rdset_t rrd
      in
      write_name root n b p;
      (* Update [nof] if [write_name] has written a pointer *)
      let nof =
	let n1 = int_of_char b.[nof] in
	if n1>63 then get_pointer b nof n1 else nof
      in 
      write_16 b p t;
      write_16 b p 1; (* the IN class *)
      write_32 b p ttl;
      write_rrd root b p nof t 1 ttl rrd)
    (m.ans @ m.aut @ m.add);
  (* edns0 *)
  (match m.edns0 with
  | Some(dob, size, rext, options) ->
      let w1 = ref 0 in
      let w2 = ref 0 in
      w1 := rext lsl 8;
      (* version = 0 *)
      if dob then setbit w2 15;
      write_null b p;  (* root *)
      write_16 b p 41; (* type = OPT *)
      write_16 b p size;
      write_16 b p !w1;
      write_16 b p !w2;
      write_options b p options
  | None -> ());
  !p, b
