(** Basic DNS definitions *)

type name
(** The abstract type of DNS names *)

val root_name : name
(** The root [name] *)

val name_of_string : string -> name
(** Get a [name] from a dotted string. *)

val string_of_name : name -> string
(** Turn a [name] into a string. *)

val labels_of_name : name -> string list
(** Splits a name into a list of strings *)

val cmp_names : name -> name -> int
(** Compare two names *)

val name_parent : name -> name
(** Return the parent name *)

val is_indomain : name -> name -> bool
(** [is_indomain a b] checks if a is in domain b *)

val is_subdomain : name -> name -> bool
(** [is_subdomain a b] checks if a is a strict subdomain of b *)

type rrt
val rrt_a     : rrt
val rrt_ns    : rrt
val rrt_cname : rrt
val rrt_soa   : rrt
val rrt_ptr   : rrt
val rrt_mx    : rrt
val rrt_txt   : rrt
val rrt_aaaa  : rrt
val rrt_opt   : rrt
val rrt_any   : rrt
val rrt_no    : rrt
val rrt_nxd   : rrt

(** The type of an RRSet *)
type rdset_t =
    A of Unix.inet_addr list
  | NS of name list
  | CNAME of name
  | SOA of name * name * int * int * int * int * int
  | PTR of name list
  | MX of (int * name) list
  | TXT of string list list
  | AAAA of Unix.inet_addr list
  | OPT of string
  | TypeOther of int * string list
  | Bad of rrt * string list

  | No of rrt * name
  | NxD of name

type class_t = IN | Class_other of int
val rrt_of_string : string -> rrt
val string_of_rrt : rrt -> string
val rrt_of_rdset_t  : rdset_t -> rrt

type rrset_t = name * int * rdset_t

type parse_error =
  | BadRdata of int
  | BadClass of int
  | BadFormat of string


(** The type of a DNS message *)
type msg = {
  adr : Unix.inet_addr;
  port : int;
  len : int;
  bad : (parse_error * string) option;
  id : int;
  opcode : int;
  qr : bool;
  aa : bool;
  tc : bool;
  rd : bool;
  ra : bool;
  ad : bool;
  cd : bool;
  rcode : int;
  edns0 : (bool * int * int * (int*string) list) option;
  qdcount : int;
  ancount : int;
  nscount : int;
  arcount : int;
  que : (name * rrt * int) list;
  ans : rrset_t list;
  aut : rrset_t list;
  add : rrset_t list;
}


val make_query : Unix.inet_addr -> int -> bool -> name -> rrt -> int -> 
  (bool * int * int * (int*string) list) option -> msg
val make_empty_answer : msg -> int -> int -> msg
val make_answer : bool -> bool -> msg -> int -> int -> 
                  rrset_t list -> rrset_t list -> rrset_t list -> string -> msg
val make_aut_answer : msg -> int -> int -> rrset_t list -> rrset_t list -> rrset_t list -> string -> msg
val make_res_answer : msg -> int -> int -> rrset_t list -> rrset_t list -> rrset_t list -> string -> msg

val make_a_query : int -> bool -> name -> rrt -> int -> int -> string
val encode_msg : msg -> int * string

val decode_msg : string -> int -> Unix.inet_addr -> int -> msg
(** [decode_msg b l adr port] decodes a [msg] from the first [l] bytes of
    buffer [b].  If l is 0, use whole buffer. *)


val string_sub : string -> int -> string
(* TODO: ZZZ remove this *)
