open Muthr
open Dns
open Retry

let debug = ref false
let dbg s = if !debug then Printf.printf "%s\n" s
let error s = Printf.fprintf stderr "%s\n" s; exit 1

(** Stub DNS resolver for muthreads *)


(*s Configuration *)

(* 

It is read from the resolv.conf file.  Currently, only the adresses of
recursive servers ('nameserver' entries) are taken into account.

 *)

let resolv_conf = "/etc/resolv.conf"

let resolvers = ref []

let max_tries = ref 3
let request_tm = ref 5.


let read_conf () =
  let f = open_in resolv_conf
  in
  let rec loop acc =
    try
      let l = input_line f in
      Scanf.sscanf l "%s %s" (fun key value ->
      loop (if key="nameserver" then (Unix.inet_addr_of_string value)::acc else acc))
    with
      End_of_file -> close_in f; acc
  in
  resolvers := List.rev (loop [])


(*s Sending and receiving DNS messages *)


(* receive a dns msg on the given socket *)
let dns_recv fd k =
  let b = String.create 512 in
  read fd b 0 512 >>= fun l -> 
  let m = decode_msg b l in
  k m

(* wait for the answer to the question *)

let wait_answer fd id question k =
  retry_forever (fun k ->
    dns_recv fd >>= fun m ->
    if m.id = id && m.que = question && (m.rcode = 0 || m.rcode = 3) then
      k (Some m)
    else
      (dbg "incorrect answer"; dbg (string_of_msg m);
      k None))
    k

(* Send a request to server and wait indefinitely for the (correct)
   answer. *)

let send_one b fd id question server k () =
  (* TODO: muthr to provide non blocking sendto? *)
  dbg (Printf.sprintf "sending request to %s"
	 (Unix.string_of_inet_addr server));
  Unix.sendto fd b 0 (String.length b) [] 
    (Unix.ADDR_INET (server, 53));
  wait_answer fd id question >>= fun r -> k (Some r)


(* Full resolution.  Send request, wait for answer, retransmit if
   necessary. Cycle through available servers, try each one ![max_tries]
   times. *)

let resolve tm qn qt qc servers k =
  let rand = Random.int 65411 in
  let id   = rand + 1024 in
  let msg  = make_a_query id qn qt qc 0 in
  let fd   = socket Unix.PF_INET Unix.SOCK_DGRAM 0 in
  let start = Unix.gettimeofday ()
  in
  retry !max_tries (fun k ->
    retry_list_tm servers tm
      (send_one msg fd id [| qn, qt, qc |])
      (fun () -> k None)
      (fun r -> k (Some r)))
    (fun () -> failwith "No answer")
    (fun m  -> k (Unix.gettimeofday() -. start, m))



(* build an [h_entry] from a DNS answer message.
   
*)
  
let h_entry_of_msg m =
  if m.rcode <> 0 then raise Not_found else
  let na = m.ancount in
  let alias = ref [] in
  let ips   = ref [] in
  let can   = ref root_name
  in
  List.iter (fun (n, c, ttl, rrset) ->
    match rrset with
    | CNAME n'    -> alias := n :: !alias
    | A (_ips) -> ips := _ips; can := n
    | _ -> raise Not_found)
    m.ans;
	  
  if !ips = [] then raise Not_found else
  { Unix.h_name = string_of_name !can;
    Unix.h_aliases = Array.of_list (List.map string_of_name !alias);
    Unix.h_addrtype = Unix.PF_INET;
    Unix.h_addr_list = Array.of_list (List.rev !ips)
  }


(** [gethostbyname] should return a [host_entry] or raise [Not_found].
   It also accepts an IPv4 or IPv6 address instead of a name.  *)
let gethostbyname n k =
  try
    let adr = Unix.inet_addr_of_string n in
    k { Unix.h_name = n;
	Unix.h_aliases = [||];
	Unix.h_addrtype = Unix.PF_INET;
	Unix.h_addr_list = [| adr |]
      }
  with Failure _ ->
    trywith
      (fun k -> resolve !request_tm (name_of_string n) rrt_a 1 !resolvers k)
      (fun _ -> raise Not_found)
      (fun (d,m) -> k (h_entry_of_msg m))


(* TODO: getaddrinfo *)
