
(* This file implements conversion of rrsets from and to text
   representations.  Recall we know no such thing as an RR, only
   rrsets.  An rrset is made of a name, a TTL, and a rdset.  We dont
   follow the master file format described in rfc1035-5.  See
   documentation for description of the format we use. TODO *)

open Dns
   

(*s Text representation *)

(* We start by defining a few shortcuts, the text representation of
   TXT rdata (we need to escape quotes and non printable characters)
   and the representation of unknown RR types, as specified in
   rfc3597. 
 *)

let id = fun a -> a

let son = string_of_name
let nos = name_of_string
let spr = Printf.sprintf
let sot t = String.escaped t  (* string of TXT rdata *)
  
let string_of_other rdata =
  let string_explode s =
    let rec l = ref [] in
    for i = 1 to String.length s do
      l := s.[i-1] :: !l
    done;
    List.rev !l
  in
  let hex = String.concat " " 
      (List.map 
	 (fun b -> Printf.sprintf "%02x" (int_of_char b))
	 (string_explode rdata))
  in
  Printf.sprintf "\\# %i %s" (String.length rdata) hex


(* [string_of_rdset] provides a multiline representation: each rdata
   is on a separate indented line.  [string_of_rdset_sl] provides a
   \emph{single line} representation, with rdatas separated by single
   spaces. TODO: review API *)

(*i ZZZ: only one space between A and %s in single line repr i*)

let string_of_rdset1 sep rr =

  let pr f l = String.concat sep (List.map f l)
  in
  match rr with
  | A ips   -> spr "A      %s" (pr Unix.string_of_inet_addr ips)
  | AAAA ips-> spr "AAAA   %s" (pr Unix.string_of_inet_addr ips)
  | CNAME n -> spr "CNAME  %s" (son n)
  | NS ns   -> spr "NS     %s" (pr son ns)
  | SOA(m,r,s,ref,ret,exp,min)  
            -> spr "SOA    %s %s %i %i %i %i %i" (son m) (son r) s ref ret exp min
  | MX ms   -> spr "MX     %s" (pr (fun (p,e) -> spr "%i %s" p (son e)) ms)
  | PTR ns  -> spr "PTR    %s" (pr son ns)
  | TXT tss -> spr "TXT    %s" (pr (fun ts -> (String.concat ", " (List.map (fun s -> "\"" ^ (sot s) ^ "\"") ts))) tss)
  | OPT s   -> spr "OPT    %s" (string_of_other s)
  | TypeOther(i, ds)
               -> spr "TYPE%i %s" i (pr string_of_other ds)
  | Bad(t, ds) -> spr "BADTYPE%i %s" (Obj.magic t) (pr string_of_other ds)
  | No(t,soa)  -> spr "! %s <%s>" (string_of_rrt t) (string_of_name soa)
  | NxD soa    -> spr "NXD <%s>" (string_of_name soa)
  | _ -> assert false


let string_of_rdset    = string_of_rdset1 "\n\t\t\t\t\t\t  "
let string_of_rdset_sl = string_of_rdset1 " "


(*s Printing DNS messages *)

let concat f l = String.concat "" (List.map f l)

let string_of_edns_options l =
  concat (fun (c,v) -> Printf.sprintf "<%i,\"%s\">" c (sot v)) l

(* The sizes of sections are given in number of \emph{rrsets} not
   \emph{RRs} as in the encoded message.
*)

let string_of_msg m =
  let son = string_of_name
  in
  let print_section name count it =    
    if count <> 0 then
     spr "%s: " name ^
    concat (fun (n, ttl, r) -> 
      spr "\t%-25s %6i   %s\n" (son n) ttl (string_of_rdset r)) it
    else ""
  in
  String.concat "" [ spr 
  "id=%i len=%i adr=%s:%i\n%s %s %s %s %s %s %s opcode=%i  \
		       rcode=%i : %i %i %i %i"
  m.id m.len (Unix.string_of_inet_addr m.adr) m.port (if m.qr then "QR" else "qr") 
  (if m.aa then "AA" else "aa") (if m.tc then "TC" else "tc") 
  (if m.rd then "RD" else "rd") (if m.ra then "RA" else "ra")
  (if m.ad then "AD" else "ad") (if m.cd then "CD" else "cd")
  m.opcode m.rcode
  (List.length m.que) (List.length m.ans) (List.length m.aut) (List.length m.add);
  (match m.edns0 with 
  | Some(dob, size, rce, options) -> 
      spr "  EDNS0: %s rce=%i size=%i %s\n" (*(if dob then "do" else "--") rce size*)
    (if dob then "DO" else "do") rce size
    (string_of_edns_options options);
  | None -> "\n");
  if m.qdcount <> 0 then spr "\nQue:" else "";
  String.concat ""
    (List.map (fun (n,t,c) -> spr "\t%-25s CLASS%i   %s\n" 
	(son n) c (string_of_rrt t)) m.que);

  print_section "Ans" m.ancount m.ans;
  print_section "Aut" m.nscount m.aut;
  print_section "Add" m.arcount m.add
]

(* Printing a msg on a single line.  TODO: use tcpdump representation?
*)

let short_string_of_msg m =
    let err = "<Not representable>"
    in
    match m.que with [ n,t,c ] ->
    let q = spr "qu: %s %s   " (son n) (string_of_rrt t)
    in q ^ (match m.ans with 
        (n,ttl,r)::_ ->
             spr "an: %s %i %s" (son n) ttl (string_of_rdset_sl r) ^
             if List.length m.ans > 1 then " ..." else ""
         | _ -> ""
         )
    | _ -> err



(*s Parsing text representation of rrsets *)

(* One rrset is represented as blank separated name, ttl, type and
   rdset. The rdset itself can be on a single line or with each rdata
   on a separate line, in which case those lines must be indented.
   (see [string_of_rdset] and [string_of_rdset_sl] above).

   The primitive function [rdset_of_string_sl] reads from a single
   line.  Other functions are provided that turn a multi line
   representation into a single line one and provide it to the
   primitive function.

   [scan_list] and [scan_single] read the rdatas for general and
   singleton only rrsets.  [scan_list] builds a list of rdatas from
   string [s] using repeatedly scanning format [fmt] and processing
   function [f], while [scan_single] does it one time only and checks
   the whole string has been consumed.

*)

let scan_list s fmt f =
  let b = Scanf.Scanning.from_string s in
  let rec loop acc =
    if Scanf.Scanning.end_of_input b then List.rev acc
    else
     loop ((Scanf.bscanf b fmt f)::acc) 
  in
  loop []

(* A TXT rdata is a (comma separated) list of string.  A TXT rdset is
   a (blank separated) sequence of TXT rdatas. 
   Here is a single rdata TXT rrset:

   1.1.1.201.aspath.routeviews.org.  81755   TXT \
       "8492 9002 2914 12956 10429 27699", "201.1.0.0", "17"

   And here is a 3 rdata TXT rrset:

   testtxt              916   TXT    "un", "deux", "trois"
                                     "quatre"
                                     "cinq", "six"
*)

let scan_txt s =
  let b = Scanf.Scanning.from_string s in
  let rec loop l acc =
    if Scanf.Scanning.end_of_input b then List.rev ((List.rev acc)::l)
    else
      match 
	try let s = Scanf.bscanf b ", %S" (fun s -> s) in Some s with _ -> None
      with Some n -> loop l (n::acc) 
	| None ->    loop (List.rev acc::l) [Scanf.bscanf b " %S" (fun s -> s)]
  in
  loop [] [Scanf.bscanf b " %S" (fun s -> s)]


let scan_single s fmt f =
  let b = Scanf.Scanning.from_string s in
  let r = Scanf.bscanf b fmt f in
  if not (Scanf.Scanning.end_of_input b) then failwith "scan_single"
  else r

let rdset_of_string_sl s =
  let t, s = Scanf.sscanf s "%s %[^\n]" (fun t s -> rrt_of_string t,s)
  in
  match t with
    | _ when t=rrt_a     -> A (scan_list s " %s" Unix.inet_addr_of_string)
    | _ when t=rrt_ns    -> NS (scan_list s " %s" name_of_string)
    | _ when t=rrt_cname -> CNAME (scan_single s " %s" name_of_string)
    | _ when t=rrt_soa   -> scan_single s "%s %s %i %i %i %i %i" 
      (fun m r s ref ret exp min -> SOA(nos m,nos r,s,ref,ret,exp,min))
    | _ when t=rrt_ptr   -> PTR(scan_list s " %s" name_of_string)
    | _ when t=rrt_mx    -> MX (scan_list s " %i %s" (fun p n -> p, name_of_string n))
      
    | _ when t=rrt_txt   -> TXT (scan_txt s)
    | _ when t=rrt_aaaa  -> AAAA(scan_list s " %s" Unix.inet_addr_of_string)
    | _ when t=rrt_no    -> scan_single s "%s <%s@>" 
      (fun t soa -> No(rrt_of_string t, name_of_string soa))
    | _ when t=rrt_nxd   -> scan_single s "<%s@> " 
      (fun soa -> NxD (name_of_string soa))
    | _ -> 
      let parse_other s =
	let l, s = Scanf.sscanf s "%i %[ 0-9a-f]" (fun i s -> (i,s)) in
	let r = String.create l in
	let bs = scan_list s " %2x " char_of_int in
	let i = ref 0 in
	List.iter (fun b -> r.[!i] <- b; incr i) bs;
	r
      in
      (try 
	 TypeOther(Obj.magic t, scan_list s " \\# %[ 0-9a-f]" parse_other)
       with _ ->  Bad(Obj.magic t, [s]))
	
    | _ -> failwith "rdset_of_rrt_sl"


(* Read a full rrset from a single line *)

let rrset_of_string_sl s =
  let n, ttl, rd = Scanf.sscanf s "%s %i %[^\n]" (fun a b c -> a,b,c)
  in
  name_of_string n, ttl, rdset_of_string_sl rd


(* We now consider the multi line representation.  We'll need to
   separate possibly several rrsets from a single multi-line string
   (in particular since we want to allow the owner name not to be
   needed if it's the same from the previous rrset).  If a line is not
   indented, it is the beginning of a new rrset.  It it is, it may be
   either the continuation of the current rrset or the beginning of a
   new rrset with the same owner name.  In the first case, the line
   starts with a new rdata, in the second one with the new rrset TTL,
   followed by the type.  Unfortunately, we can't just check if we see
   the TTL, since rdatas sometimes start with an int (eg MX).
   [is_new_rrset] checks if a line starts with an int (could be a TTL)
   followed by (the text representation of) an [rrt].

*)

let is_indented l =
  let c = l.[0] in c = ' ' or c = '\t'

let is_new_rrset s =
  try
    Scanf.sscanf s " %i@ %s@ " (fun i s -> let _ = rrt_of_string s in true)
  with _ -> false

(* This one could/should exist in a general purpose library. *)
let string_split sep s =
  let rec help s acc =
    try
      let i = String.index s sep in
      help (string_sub s (i+1)) (String.sub s 0 i :: acc)
    with Not_found ->
      List.rev (s :: acc)
  in
  help s []

(* From a string [s] containing the text representation of an unknown
   number of rrsets, return a list of strings, one for each rrset. *)

let split_rrsets s sep =
  let ss = string_split '\n' s 
  in
  let c, r = List.fold_left (fun (cur,acc) l -> 
    if l = "" then ("", cur::acc) else
    if is_indented l && not (is_new_rrset l) then ((cur ^ sep ^ l), acc)
    else (l, cur::acc))
    ("", [])
    ss
  in List.rev (c::r)

let split_lines s = split_rrsets s ""

(* We allow for empty lines and comment lines starting with the ';'
   character. *)

let rrsets_of_string s =
  let rec loop ls acc pname =
    match ls with
    | l::ls ->
	if l = "" || l.[0] = ';' then
	  loop ls acc pname
	else
          let (pname,_,_) as nr = rrset_of_string_sl 
            (if is_indented l then (string_of_name pname) ^ " " ^ l else l)
          in
          loop ls (nr::acc) pname

    | [] -> acc
  in
  loop (split_lines s) [] root_name
