(*s Main dong file *)

open Muthr
open Dns
open Dns_txt
open Answer
module L = List
open Vars

let version = "dong v 0.51"

(* Tunable parameters *)

let log_file   = ref "/var/log/dong.log"

let hints_file = ref "/home/deleuzec/Rech/Muthreads/Dns/hints.dong"
let nsid       = ref "default nsid string"

let load_cache  = ref true  (* load/save cache to file *)
let rrsets_file = ref "/var/cache/dong.rrsets"
let auths_file  = ref "/var/cache/dong.auths"

let save_bad   = ref false  (* save dump of bad queries/replies *)
let sb_file    = ref "/var/cache/dong.bad"


let quota        = ref 8 (* to bound thread work *)

let  purge_itv   = ref 300.
let spurge_itv   = ref 301.
let mng_port     = ref 9000

(*let (sp, cf) = Bad_msg.open_file !sb_file*)
let (sp, cf) = (fun _ _ -> ()), fun () -> ()

(* Statistics *)

let start_time = ref 0.

let cur_rsv = ref 0  (* nb of current resolutions *)

let sro = ref 0  (* resolutions with ok result *)
let srf = ref 0  (* resolutions failed *)
let srF = ref 0  (* resolutions crashed -- should always be 0! *)

let sre = ref 0  (* resolutions executed *)
let sra = ref 0  (* resolutions aggregated *)
let src = ref 0  (* resolutions that found result in cache *)


let make_answer m rcode ms ans aut add = make_res_answer m rcode ms ans aut add !nsid


(*s Resolution *)

(* RFC1034-5.3.3. describes the resolver algorithm as a 4 steps
   procedure:

   \begin{enumerate}
   \item lookup cache,
   \item find best servers,
   \item send request until acceptable answer,
   \item analyze the response.
   \end{enumerate}

  Below, [resolv1] starts resolution at step 1, while [resolv2] starts
  at step 2.  Before to engage in a resolution, we check if a
  resolution is already running for the same question.  If so, we'll
  just wait for its result.  
*)

let dbg_start ste quota =
  dbg (Printf.sprintf "T:+ .%i. %s %s (q=%i)"
	 ste.id (string_of_name ste.sn) (string_of_rrt ste.st) quota)  

let dbg_res ste cnames res =
  dbg (Printf.sprintf "T:- .%i. %s %s: %s %s" 
	 ste.id (string_of_name ste.sn) (string_of_rrt ste.st) (string_of_answer_t res)
	 (if cnames = [] then "" else "Cnames: " ^ (String.concat "" (List.map (fun (t,r) -> 
	   Printf.sprintf "%i %s\n" t (string_of_rdset r)) cnames))))


(* ZZZ to be cleaned up! *)
let ans_of_cache (ttl,rrs) =
  match rrs with
  | No(t,soa) -> NoData(ttl,soa)
  | NxD soa   -> NxDomain(ttl, soa)
  | _         -> Data(ttl, rrs)

let rec resolv iquota ste k =
  let quota = ref iquota
  in
  let on = ste.sn in (* save initial sname *)
  incr cur_rsv;
  dbg_start ste iquota;

let rec resolv1 acc ste k =
  dbg_st "resolv1" ste "";

  match Cache.get_follow ste.sn ste.st with
  | qname, cnames, None -> incr sre;
      ste.sn <- qname;
      let zone, q = Cache.find_auth ste.sn in
      resolv2 (cnames @ acc) ste zone q k

  (* was in cache, we're done! *)
  | qname, cnames, Some res -> incr src; k (cnames @ acc, ans_of_cache res)

and resolv2 acc ste zone q k =
  dbg_st "resolv2" ste ("q=" ^ string_of_int !quota);

  decr quota;
  if !quota<=0 then k (acc, RFailed "Exceeded quota")
  else

  Resolv.query_multiple_servers (resolv (iquota-1)) ste zone q >>= fun (ncns, res) ->
  List.iter Cache.add_rrset ncns;
  let acc = (List.map (fun (_,t,d) -> t,d) ncns) @ acc
  in
  match res with

  (* Partial answer *)
  | Referral(qn, ns, add) -> let (z,t,q) = make_auth ns add in
    Cache.add_auth (z,t,q); resolv2 acc ste z q k

  | Alias n -> ste.sn <- n; resolv1 acc ste k

  (* Final answer *) (* ZZZ incr sro *)
  | Data (t, rds) -> Cache.add_rrset (ste.sn, t, rds); k (acc, res)
  | DataAny _     -> k (acc, res)
  | NoData(t,n)   -> Cache.add_rrset (ste.sn, t, No(ste.st,n)); k (acc, res)
  | NxDomain(t,n) -> Cache.add_rrset (ste.sn, t, NxD n); k (acc, res)
  | RFailed _     -> incr srf; k (acc, res)

  (* Impossible answer (!) *)
  | Invalid _
  | Lame _
  | Truncated
  | Error _    -> failwith "There's something wrong. This can't happen!"

in

(* Check ongoing resolutions table *)
  Running.check_running ste.sn ste.st ste.id ste.parents >>= fun r ->
  match r with
  | Running.GotIt (cnames, res) ->
      incr sra;
      decr cur_rsv;
      k (cnames, res)

  | Running.DoIt handle ->
      trywith
	(resolv1 [] ste)

        (* This should not happen *)
	(fun e k  -> incr srF; k ([], (RFailed ("ARG! " ^ Printexc.to_string e))))

	(fun (cnames, res) -> 
      ste.sn <- on;  (* put back original sname for logs *)
      dbg_res ste cnames res;
      decr cur_rsv;
      handle (cnames, res) >>= k
      )
	  

(*s Building final answers *)

(* Build a [rrs_t list] to put in answer section of a reply message
   from the final answer, its ttl and the cnames. *)

let make_chain qn ttl rrs cnames =
  if List.length cnames = 0 then
    [ qn, ttl, rrs ]
  else
    let (c,r) = List.fold_left
	(fun (q,list) (ttl,(CNAME c)) -> (c,(q,ttl,CNAME c)::list))
	(qn,[])
	(List.rev cnames)
    in
    List.rev ((c,ttl,rrs)::r)

let make_chain_nodata qn cnames =
  if List.length cnames = 0 then
    [ ]
  else
    let (c,r) = List.fold_left
	(fun (q,list) (ttl,(CNAME c)) -> (c,(q,ttl,CNAME c)::list))
	(qn,[])
	(List.rev cnames)
    in
    List.rev r

let make_chain_any qn l cnames =
  let (c,r) = List.fold_left
      (fun (q,list) (ttl,(CNAME c)) -> (c,(q,ttl,CNAME c)::list))
      (qn,[])
      (List.rev cnames)
  in
  List.rev r @ List.map (fun (t,r) -> c, t, r) l


let build_soa n ttl =
  try
    match Cache.get_rrset n rrt_soa with 
    | _, soa -> [ n, ttl, soa ]
    | _ -> []
  with Not_found -> []

(* send reply.

   TODO: fill in authority and additional sections

 *)

let send_reply s m qn qt qc cnames res =
  
  let m = match res with
  | Data(ttl, rrs) ->
      make_answer m 0 msg_size (make_chain qn ttl rrs cnames) [] []
	(* qname ? *)
	
  | DataAny l ->
      make_answer m 0 msg_size (make_chain_any qn l cnames) [] []

  | NoData(ttl,n) ->
      make_answer m 0 msg_size 
	(make_chain_nodata qn cnames) (build_soa n ttl) []

  (* ZZZ find an example of NxDomain with a cname? *)
  | NxDomain(ttl,n) ->
      make_answer m 3 msg_size
	(make_chain_nodata qn cnames) (build_soa n ttl) []

  | RFailed s -> make_empty_answer m 2 msg_size

  | Error n -> make_answer m n msg_size [] [] []

  | _ -> failwith ("send_reply:" ^ string_of_answer_t res)
  in
  dbg_msg "==> sending reply" m;
  dns_send_reply s m; ()


(*s Main threads *)

let resolv_thread s m () =
  if m.qdcount <> 1 then begin
    incr sbq;
    dbg "Unimplemented: qdcount <> 1";
    sp false m;
    let m = make_empty_answer m 4 msg_size in
    dns_send_reply s m; terminate ()
  end
  else
    let [ qn, qt, qc ] = m.que in
    if not m.rd || qc<>1 then
      (incr sbq; send_reply s m qn qt qc [] (Error 5);
       terminate ()
      )
    else (
      incr ser;
      resolv !quota (ext_state qn qt) >>= fun (cnames, res) ->
      send_reply s m qn qt qc cnames res; terminate())
    
(*i BUG
	    (fun k () ->
	      resolv1 [] qn qt qc >>=	fun cnames res ->
	       gotit handle (cnames, res))

does not compile
	    (fun k () ->
	      resolv1 [] qn qt qc >>=	fun cnames res ->
	       gotit handle (cnames, res); k())

i*)

let bind_loopback s port =
  Unix.bind s (Unix.ADDR_INET (Unix.inet_addr_loopback, port))

let rec main s () =
  dns_recv s >>= fun m ->
  if m.bad <> None then begin
    incr sbq;
    dbg "<=== received BAD query";
    sp false m;
  end
  else begin
    (* here we could detect a retransmitted query *)
    dbg_msg "<== received query" m;
    spawn (resolv_thread s m)
  end;
  incr srq;
  main s ()

(*s [purge] and [mng] *)

let rec purge () =
  sleep !purge_itv >>= fun () ->
  log "P: starting cache purge";
  ignore (Cache.purge ());
  log "P: cache purge done";
  purge ()

let rec spurge () =
  sleep !spurge_itv >>= fun () ->
  log "P: starting scache purge";
  Cache.spurge ();
  log "P: scache purge done";
  spurge ()

let show_stats () =
  let pct a b = float_of_int a /. float_of_int b *. 100. in
  let tot = !ser + !sir
  in
  let queries = Printf.sprintf "\
    \t received queries %5i   sent answers     %5i  %6.2f%%    bad queries %5i  %6.2f%%\n\
    \t sent queries     %5i   received answers %5i  %6.2f%%    bad answers %5i  %6.2f%%\n"
    !srq !ssr (pct !ssr !srq)  !sbq (pct !sbq !srq)
    !ssq !srr (pct !srr !ssq)  !sbr (pct !sbr !srr)
  in
  let resolv = Printf.sprintf "\
    \t external  %4i  %6.2f%%  internal   %4i  %6.2f%%\n\
    \t result ok %4i  %6.2f%%  failed     %4i  %6.2f%%   crashed    %4i  %6.2f%%\n\
    \t executed  %4i  %6.2f%%  aggregated %4i  %6.2f%%   from cache %4i  %6.2f%%\n\
    \t running   %4i           distinct   %4i\n"
    !ser  (pct !ser tot) !sir (pct !sir tot)
    (!sro+ !src)  (pct (!sro+ !src) tot) !srf (pct !srf tot)  !srF (pct !srF tot)
    !sre  (pct !sre tot) !sra (pct !sra tot)  !src (pct !src tot)
    !cur_rsv (let _,n = Running.dump_rslt () in n)
  in
  let caches = 
    let _, nn, nr = Cache.dump_all ()
    and _, na     = Cache.dump_nscache () 
    and _, ns     = Cache.dump_scache () in
    Printf.sprintf "\
    \t rrsets  %5i      names   %5i  (%3.2f rrsets/name)\n\
    \t auths   %5i      servers %5i\n"
      nr nn (float_of_int nr /. float_of_int nn)  na  ns
  in
  Printf.sprintf "Started on %s (%i seconds ago)\n\
    messages\n\
    %s\
    resolutions\n\
    %s\
    caches\n\
    %s\
    config\n\
    \t log msgs %3s\n"
    (Misc.string_of_date !start_time) (int_of_float (Unix.time() -. !start_time))
    queries
    resolv
    caches
    (if !debug_msg then "on" else "off")

let write_str s st k =
  really_write s (st ^ "\n") 0 (String.length st + 1) >>= k

let write_prompt s k =
  really_write s "dong> " 0 6 >>= k

let unconcat str sep =
  let l = String.length str
  in
  let rec loop i =
    if i=l+1 then []
    else
      let j = try String.index_from str i sep with Not_found -> l in
      if j>i then String.sub str i (j-i) :: loop (j+1) else loop (j+1)
  in
  loop 0


let write_file file st =
  let f = open_out file in
  output_string f st;
  close_out f

(* The [mng] thread allows a remote connection to set and observe
   various parameters of the running resolver.  Useful for
   debugging. *)

let mng_help = 
  "help for " ^ version ^ " mng command line\n" ^
  "dr           dump running resolutions\n\
   dc [name]    dump cached rrsets (under name)\n\
   sc <name>    show rrsets at given name\n\
   scf <name>   show rrsets at given name and related ones\n\
   flush <name> forget about rrsets name and below\n\
   ds           dump server cache\n\
   dn [name]    dump ns cache\n\
   dm [on|off]  log message contents\n\
   ss           show stats\n\
   purge        purge rrsets and server caches\n\
   save rrsets|auths <file>  save rrsets or auths caches to file\n\
   load rrsets|auths <file>  load rrsets or auths caches from file\n\
   help         display this message\n\
   exit         do that\n"

let mng () =
  let s = socket Unix.PF_INET Unix.SOCK_STREAM 0 in
  let b = String.create msg_size in
  Unix.setsockopt s Unix.SO_REUSEADDR true;
  Unix.bind s (Unix.ADDR_INET (Unix.inet_addr_loopback, !mng_port));
  Unix.listen s 1;

  let rec loop s last k =
    let nos = name_of_string 
    in
    let one_nb cmd fct arg fmt =
      let r, n = fct arg in
      write_str s (r ^ Printf.sprintf fmt n) >>= fun () -> loop s cmd k
    in
    let two_nb cmd fct arg fmt =
      let r, n, e = fct arg in
      write_str s (r ^ Printf.sprintf fmt n e) >>= fun () -> loop s cmd k
    in
    let one_nb2 cmd fct arg fmt = one_nb cmd (fun arg -> "", fct arg) arg fmt
    in

    let rec process cmd =
    match unconcat cmd ' ' with
    | [ "dr" ] -> two_nb cmd (fun () -> let r, n = Running.dump_rslt () in r,!cur_rsv,n) ()
	  "\n:%i current resolutions (%i distincts)\n"

    | [ "dc" ]      -> two_nb cmd Cache.dump_all    ()        ":%i names, %i rrsets\n"
    | [ "dc"; arg ] -> two_nb cmd Cache.dump_under  (nos arg) ":%i names, %i rrsets\n"
    | [ "sc"; arg ] -> two_nb cmd Cache.dump_one    (nos arg) ":%i names, %i rrsets\n"
    | [ "scf"; arg ]-> two_nb cmd Cache.dump_follow (nos arg) ":%i names, %i rrsets\n"

    | [ "flush"; arg ] -> two_nb cmd
	  (fun a -> let (n,r) = Cache.flush_cache a in "",n,r) (nos arg) ":%i names, %i rrsets\n"

    | [ "ds" ]      -> one_nb cmd Cache.dump_scache  ()            "\n:%i servers\n"
    | [ "dn" ]      -> one_nb cmd Cache.dump_nscache ()            "\n:%i names\n"
    | [ "dn"; arg ] -> one_nb cmd Cache.dump_nscache_all (nos arg) "\n:%i names\n"

    | [ "sn"; arg ] -> let r = Cache.dump_nscache_one (nos arg) in
      write_str s r >>= fun () -> loop s cmd k

    | ["dm"; "on" ] -> debug_msg := true;  loop s cmd k
    | ["dm"; "off"] -> debug_msg := false; loop s cmd k

    | [ "ss" ] -> let r = show_stats () in
      write_str s r >>= fun () -> loop s cmd k

    | ["help"] -> write_str s mng_help >>= fun () -> loop s cmd k
    | ["exit"] -> write_str s "Ciao"   >>= fun () -> close s; k ()

    | ["purge"] -> two_nb cmd (fun () -> Cache.spurge (); let (n,r) = Cache.purge () in "",r,n)
	  () ":%i rrsets, affecting %i names (no info about scache)\n"

    | ["save"; "auths";  file ] -> one_nb2 cmd Cache.save_auths  file ":%i auths saved\n"
    | ["load"; "auths";  file ] -> one_nb2 cmd Cache.load_auths  file ":%i auths loaded\n"
    | ["save"; "rrsets"; file ] -> one_nb2 cmd Cache.save_rrsets file ":%i rrsets saved\n"
    | ["load"; "rrsets"; file ] -> one_nb2 cmd Cache.load_rrsets file ":%i rrsets loaded\n"

    | [] -> loop s cmd k

    | ["!"] -> process last

    | _ -> write_str s "Unknown command" >>= fun () -> loop s cmd k
    in
    write_prompt s >>= fun () ->
    read s b 0 msg_size >>= fun l ->
    if l = 0 then (close s; k ()) else
    process (String.sub b 0 (l-1))
  in

  let rec session () =
    accept s >>= fun (s,adr) -> 
    write_str s "Welcome to dong caching recursive DNS server" >>= fun () ->
    Unix.setsockopt s Unix.SO_REUSEADDR true;
    loop s "" >>= session
  in
  session ()

(*s Root hints *)

(* As per rfc1034-5.3.2, we need a "safety belt" that provides servers
   to query when we know nothing better.  This is the NS rrset for the
   root and associated glue.  We normally read this from a conf file,
   but use the following builtins if this fails.  This is the initial
   content of our nscache.

   ZZZ we should perform a "priming" query to make it autoritative... ?
*)

let roots = [ 
  "a.root-servers.net.", "198.41.0.4"    , "2001:503:BA3E::2:30";
  "b.root-servers.net.", "192.228.79.201", "" ; 
  "c.root-servers.net.", "192.33.4.12"   , "" ;
  "d.root-servers.net.", "128.8.10.90"   , "" ;
  "e.root-servers.net.", "192.203.230.10", "" ;
  "f.root-servers.net.", "192.5.5.241"   , "2001:500:2F::F";
  "g.root-servers.net.", "192.112.36.4"  , "" ;
  "h.root-servers.net.", "128.63.2.53"   , "2001:500:1::803F:235";
  "i.root-servers.net.", "192.36.148.17" , "2001:7FE::53";
  "j.root-servers.net.", "192.58.128.30" , "2001:503:C27::2:30";
  "k.root-servers.net.", "193.0.14.129"  , "2001:7FD::1";
  "l.root-servers.net.", "199.7.83.42"   , "2001:500:3::42";
  "m.root-servers.net.", "202.12.27.33"  , "2001:DC3::35"
]

let set_cache_root_builtin () =
  let nss = List.map name_of_string (List.map (fun (n,_,_) -> n) roots) in
  let au = (root_name, 36000000, NS nss)
  in
  let ad = List.map 
      (fun (n,ip,ip6) -> 
	let ip = Unix.inet_addr_of_string ip in
	(name_of_string n, 360000000, A [ ip ])
      )
      roots
  in
  Cache.add_auth (make_auth au ad)

let load_rrsets f =
  let rec loop acc =
    try
      loop (input_line f :: acc)
    with End_of_file -> String.concat "\n" (List.rev acc)
  in
  rrsets_of_string (loop [])

let load_root file =
  let f = open_in file  in
  let rs = load_rrsets f in
  close_in f;
  let [au], add = List.partition (fun (n,t,rd) -> rrt_of_rdset_t rd = rrt_ns) rs
  in let add = List.filter (fun (_,_,rd) -> rrt_of_rdset_t rd = rrt_a) add
  in 
  Cache.add_auth (make_auth au add)

let set_cache_root () =
  try
    load_root !hints_file
  with _ -> 
    Cache.cache_reset ();
    log "Warning: error loading hints file, using builtin hints";
    set_cache_root_builtin ()

let cache_load () =
  let n = Cache.load_rrsets !rrsets_file in
  let m = Cache.load_auths  !auths_file  in
  if m=0 then set_cache_root ()

let cache_save () =
  ignore (Cache.save_rrsets !rrsets_file);
  ignore (Cache.save_auths  !auths_file)


(*s Start up code *)

let resolver () =
  let fd = socket Unix.PF_INET Unix.SOCK_DGRAM 0 in
  bind_loopback fd 53;
  Misc.open_log !log_file;
  set_cache_root ();
  if !load_cache then cache_load ();
  spawn mng;
  spawn purge;
  spawn spurge;
  spawn (main fd)
;;


let spec = [
	    ("-m", Arg.Set_int    max_tries,  "number of tries");
	    ("-tm", Arg.Set_float  request_tm, "retransmission timeout");
	    ("-g", Arg.Set debug, "debug");
	    ("-gm", Arg.Set debug_msg, "debug messages");
            ("-h", Arg.Set_string hints_file, "set hints file");
            ("-l", Arg.Set_string log_file, "set log file");
            ("-nsid", Arg.Set_string nsid, "set NSID string");
            ("--no-load", Arg.Clear load_cache, "do not load cache file")
	  ]
let usage_msg = "dong [options]"

let parse_args () =
  Arg.parse spec
    (fun n -> print_string "unknown arg\n";
      Arg.usage spec usage_msg; exit 1)
    usage_msg

let die n =
  Misc.log ("Received signal " ^ string_of_int n);
  if !load_cache then cache_save ();
  Misc.close_log ();
  exit 0

let init () =
  Random.self_init();
  Sys.set_signal Sys.sigterm (Sys.Signal_handle die);
  Sys.set_signal Sys.sigquit (Sys.Signal_handle die);
  Sys.set_signal Sys.sigint  (Sys.Signal_handle die);
  start_time := Unix.time();
  parse_args ();
  resolver ();
  set_debug true;
  start ()
;;

init ()
