open Dns
open Unix

(* doug: a DNS redirector

   doug waits for DNS queries for a specified set of names.
   It gives an answer based on the querier IP address.

*)

let ($) f e = f e
let dbg s = prerr_string s; flush Pervasives.stderr

(* Reading the data structure from a text file *)

(* converting a string reprensentation of a pattern (a.b/nn) in a
   (inet_addr, mask) pair *)

let count_dots st = 
  let rec loop n i =
    try
      loop (n+1) (String.index_from st (i+1) '.')
    with _ -> n
  in loop 0 0

(* complete the string prefix to have a full a.b.c.d string address *)
let complete_prefix st p =
  let rec loop n =
    if n = 0 then "" else ".0" ^ (loop (n-1))
  in
  String.sub st 0 p ^ loop (3 - count_dots st)

(* inet_addr -> string -> int32, isn't it *very* ugly?  no choice
   since inet_addr is abstract and of/to string are the only
   conversion functions available

   (we could add to/of Int32 functions in unix.ml? )

   However, this in only done at start up when parsing the conf
   file...
*)

let int32_of_inet_string st =
  let b1 = String.index st '.' in
  let b2 = String.index_from st (b1+1) '.' in
  let b3 = String.index_from st (b2+1) '.'
  in
  let byte b e = Int32.of_int (int_of_string (String.sub st b (e-b)))
  in
  let (+.) = Int32.add
  in
  Int32.shift_left (byte 0 b1) 24 +.
  Int32.shift_left (byte (b1+1) b2) 16 +.
  Int32.shift_left (byte (b2+1) b3) 8 +.
  byte (b3+1) (String.length st)

let int32_of_inet_addr addr =
  int32_of_inet_string (Unix.string_of_inet_addr addr)

let string_of_int32 i =
  let toi i32 = let v = Int32.to_int i32 in if v < 0 then v+256 else v
  in
  let b1 = Int32.shift_right i 24
  and b2 = Int32.logand (Int32.shift_right i 16) (Int32.of_int 255)
  and b3 = Int32.logand (Int32.shift_right i 8) (Int32.of_int 255)
  and b4 = Int32.logand i (Int32.of_int 255)
  in
  Printf.sprintf "%i.%i.%i.%i"  (toi b1) (toi b2) (toi b3) (toi b4)

(* Int32.shift_left only specified for n <=31 ! *)
let shift v n =
  if n<32 then Int32.shift_left v n
  else Int32.shift_left (Int32.shift_left v 1) 31

(* Finally, get (ip, mask) as int32s from string pattern *)

let pattern_of_string st =
  let p  = String.index st '/' in
  let n  = int_of_string (String.sub st (p+1) (String.length st - p - 1)) in
  let ip = int32_of_inet_string (complete_prefix st p)  in
  let mask = shift (Int32.lognot Int32.zero) (32-n) in
  let ip = (Int32.logand ip mask)
  in

  (* just to make sure!*)
  Printf.printf "prefix %s -> ip %s, mask %s\n" st
    (string_of_int32 ip) (string_of_int32 mask);
  flush Pervasives.stdout;

  ip, mask

(* format is
   name prefix TTL rdset
        prefix TTL rdset
        ...
   name ...
*)

let scan_list b fmt f =
  let rec loop acc =
    if Scanf.Scanning.end_of_input b then List.rev acc
    else
     loop ((Scanf.bscanf b fmt f)::acc) 
  in
  loop []

let vrrset_of_string s =
  let b = Scanf.Scanning.from_string s in
  let n = Scanf.bscanf b "%s " (fun n -> name_of_string n) in
  n, scan_list b " %s %i %[^\n]"
    (fun pf ttl rd -> pattern_of_string pf, ttl,  Dns_txt.rdset_of_string_sl rd)

let is_indented l =
  let c = l.[0] in c = ' ' or c = '\t'

let vrrsets_of_string s =
  let rec loop ls acc pname =
    match ls with
    | l::ls ->
	if l = "" || l.[0] = ';' then
	  loop ls acc pname
	else
          let (pname,_) as nr = vrrset_of_string
            (if is_indented l then (string_of_name pname) ^ " " ^ l else l)
          in
          loop ls (nr::acc) pname

    | [] -> acc
  in
  loop (Dns_txt.split_rrsets s "\n") [] root_name

let readFile f =
  let f = open_in f in
  let l = in_channel_length f in
  let b = String.create l in
  really_input f b 0 l;
  close_in f;
  b

let data = vrrsets_of_string (readFile "doug.conf")
  

(* Does src match (ip, mask)? *)

let do_match (ip, mask) src =
  let src = int32_of_inet_addr src
  in
  dbg $ Printf.sprintf "trying ip=%s pf=%s msk=%s\n" (string_of_int32 src) (string_of_int32 ip) (string_of_int32 mask);
  (Int32.logand src mask) = ip


(* The server itself *)

let debug_msg = ref false
let dbg_msg t m =
  if !debug_msg then Printf.printf "\n%s\n%s" t (Dns_txt.string_of_msg m);
  flush Pervasives.stdout

let msg_size = 512

let dns_recv fd =
  let b = String.create msg_size in
  let (l,Unix.ADDR_INET(adr,port)) = recvfrom fd b 0 msg_size []
  in
  decode_msg b l adr port
  
let dns_send fd m =
  let l, b = encode_msg m in
  Unix.sendto fd b 0 l [] 
    (Unix.ADDR_INET (m.adr, m.port))

let bind_any s port =
  Unix.bind s (Unix.ADDR_INET (Unix.inet_addr_any, port))


let get_match src_ip list =
  let rec loop l =
    match l with
    | [] -> raise Not_found
    | (p, ttl, rrset) :: t ->
	if do_match p src_ip then (ttl,rrset)
	else loop t
  in
  loop list

(* If the message contains several questions or a question with qtype
   not A or qclass not IN, we reply with REFUSED (rcode=5).

   If we don't know about the name or the source ip doesn't match any
   prefix, we reply with NxDOMAIN (rcode=3).

   Otherwise, we reply NOERROR with the found rrset.
*)

let make_answer m rcode ans aut add =
  Dns.make_answer true false m rcode 0 ans aut add ""

let process_request m =
  dbg_msg "<< " m;
  try
    let [qn,qt,qc] = m.que in
    if qt<>rrt_a or qc<>1 then failwith "bad_question"
    else
      try
	let entry = List.assoc qn data in
	let ttl,rrset = get_match m.adr entry 
	in
	make_answer m 0 [ qn, ttl, rrset ] [] []
      with 
	Not_found -> make_answer m 3 [] [] []
  with _ ->
    make_answer m 5 [] [] []

    
let main () =
  let fd = socket Unix.PF_INET Unix.SOCK_DGRAM 0 in

  let rec loop () =
    let m = dns_recv fd in
    let r = process_request m in 
    dbg_msg ">> " r;
    dns_send fd r;
    loop ()
  in
  bind_any fd 53;
  loop ()
;;

main ()
