open Muthr
open Muthr_retry
open Vars
open Dns
open Answer

(*s One step of resolution *)

(* The resolution is made of several steps: at each one, we query
   servers that have authority on a zone.  They may give us a final
   answer if they do have authority on the queried name, or a referral
   to a zone closer to the queried name (along with the authoritative
   servers' names for that zone).

   Thus, on each such step, we want to query authoritative servers for
   the current zone. We may have to query several of them if we don't
   receive timely and correct answers.  We'll do that by spawning
   several threads.

*)

let timed_recv d fd kfail k =
  timeout d (fun k -> dns_recv fd k) kfail k

let dbg_ros sname stype (acc,res) =
  dbg ("ros: " ^ (string_of_name sname) ^ " " ^ (string_of_rrt stype) ^
       ": " ^ (string_of_answer_t res) ^
       (if acc=[] then "" else " cnames " ^ 
       (String.concat " " 
	  (List.map (fun (_, t, CNAME n) -> string_of_name n) acc))))

(*s One step of resolution: querying one server *)

(* Let's first look at the code for querying one server.  [sn], [st],
  [sc] describe the query.  [zone] is the zone the server [srv]
  (either an ip address or a name)has authority on.  [get_ip] is for
  getting the ip address from [srv] (running an internal resolution if
  needed).  [succeeds] and [fails] are to be called for the
  corresponding outcome.

  Querying one server consists of building and sending the query and
  waiting for a reply, retransmitting the query after a timeout.
  We'll use the [retry] muthread function for that.

  But EDNS complicates things a bit. RFC2671-5.3 says that
  ``Responders who do not understand these protocol extensions are
  expected to send a response with RCODE NOTIMPL, FORMERR, or
  SERVFAIL''.  Thus, a thread will hold and update local state of type
  [edns_state] to know what to do about EDNS: do not use it, try it,
  try not to use it (ie, we tried it and it failed.)

 *)

type edns_state =
  | NoEdns | TryEdns | TryNoEdns

let query_one_server ste zone srv get_ip succeeds fails () =
  get_ip srv >>= fun ip ->
  let (edns, is_lame) = Cache.scache_get ip zone in
  if is_lame then fails () else
  let edns = ref (if edns then TryEdns else NoEdns) in
  let fd = socket Unix.PF_INET Unix.SOCK_DGRAM 0 in
  let id = Random.int 64511 + 1024
  in

(*

   If we receive an answer other than [Error] and [Invalid] to a query
   without EDNS after one with EDNS that failed ([edns] current value
   is [TryNoEdns]), we cache the server's unsupport of EDNS.

   *)

  let update_edns_support () =
    if !edns=TryNoEdns then Cache.scache_add_noedns ip
  in

  let examine_answer m =
    dbg_msg "<<<" m;
    let cnames, res = Analyze.process_answer ip id ste.sn ste.st ste.sc zone m in
    match res with

(*
  If we receive an [Error], then if [edns] refers to [TryEdns] our
  query was using EDNS, we change [edns] to [TryNoEdns] and retry the
  query.  If not, the error is not due to EDNS unsupport, we fail
  normally.
*)

    | Error i -> 
	if !edns = TryEdns then (edns := TryNoEdns; None)
	else (dbg_msg2 ("<<< *** Error " ^ string_of_int i); raise Failed)

    | Truncated -> dbg_msg2 "*** Truncated"; 
	update_edns_support (); raise Failed
	  
    | Lame _ -> dbg_msg2 "*** Lame";
	Cache.add_lame ip zone; 
        update_edns_support (); raise Failed
	  
    | Invalid s -> incr sbr;
	dbg_msg2 ("*** Invalid: " ^ s); raise Failed

    | _ -> dbg_ros ste.sn ste.st (cnames, res);
	update_edns_support ();
	Some(cnames, res)
  in

  retry !max_tries
    (fun k ->
      let b  = make_a_query id false ste.sn ste.st ste.sc 
	  (if !edns=TryEdns then msg_size else 0)
      in
      dbg_msg (">>> sending query for zone " ^ (string_of_name zone))
	(decode_msg b 0 ip 53);
      dns_send_raw fd ip b;
      
      timed_recv !request_tm fd
	(fun () -> dbg_st "retry" ste 
	    (Printf.sprintf " on %s" (Unix.string_of_inet_addr ip)); k None)
	(fun m ->  incr srr; k (examine_answer m)))

    (fun () -> close fd; fails ())
    (fun r  -> close fd; succeeds r)

(*s One step of resolution: querying multiple servers *)

(* Here is the code for querying multiple servers.

  [stop], [finish], [cnt] and [next] are for coordination amongst
  concurrent threads querying servers for the same zone.  [stop] is a
  shared boolean ref which will be set by one thread if it succeeded,
  so that all other threads don't perform any more useless work.

  [finish] is an mvar where the final result will be put.  This will
  be done either by the first suceeding thread, or by the last one to
  finish if all have failed.  The thread will know it's the last one
  because [cnt], the number of threads still to be completed, will be
  zero.  Each failing thread decrements it.

  The [loop] thread ZZZ ends as soon as one server thread has
  succeeded ([stop] has been set) or when it has no more server to
  consider.

 *)

let shuffle l =
  let rec loop n pre suf =
    if n = 0 then suf @ pre (* should we List.rev pre ? *)
    else
      match suf with
      | h::t -> loop (n-1) (h::pre) t
      | _ -> loop 0 pre suf
  in
  if l = [] then [] else
  loop (Random.int (List.length l)) [] l

let query_multiple_servers rsv ste zone l k =
  let stop = ref false in
  let cnt = ref (List.length l) in
  let finish = make_mvar ()
  in

  let succeeds (cn,res) = 
    (if !stop then skip else (stop:=true; put_mvar finish (cn,res))) >>=
     terminate
  in
  let fails next () = 
    decr cnt;
    (if !cnt = 0 then
      let res =  ([], RFailed ("failed at zone " ^ string_of_name zone)) in
      put_mvar finish res
    else
      put_mvar next ()
    ) >>= terminate
  in

  let get_ip next s k =
    let subresolv rsv n k =
      incr sir;
      rsv (int_state n rrt_a ste) >>= fun (_, res) ->
      match res with
      | Data(_, A (ip::_)) -> k (Some ip)
      | _ -> k None
    in
    (match s with
    | Answer.Extern name -> subresolv rsv name
    | Answer.Glue ip -> fun k -> k (Some ip)) >>= fun r ->
    match r with
    | None -> fails next ()  (* We couldn't resolve the [Extern] name. *)
    | Some ip -> 
	(* Since we may have blocked, check again
	   if some other thread has succeeded *)
	if !stop then terminate () else k ip
  in

  let rec loop l () =
    if !stop then terminate () else match l with
    | [] -> terminate ()
    | srv::t -> 
	let next = make_mvar ()
	in
	spawn (query_one_server ste zone srv (get_ip next) succeeds (fails next));
	timed_take !mretry_delay next >>= fun _ -> 
	loop t ()
  in
  spawn (loop (shuffle l));
  take_mvar finish >>= k

