
(* *)

open Muthr
open Dns
open Answer
module Misc = Muthr_misc

let debug = ref true

let dbg s = if !debug then Misc.log s


(*

The complete result of a resolution has type [res_t]: the [answer_t]
along a list of CNAMEs.  The result of a table lookup will be:

\begin{itemize}

\item[DoIt handle]: please perform the resolution, then apply handle
  to the result,

\item[GotIt res]: here's the result (you have blocked, though you
  didn't realize it).

\end{itemize}
*)

type res_t = (int * rdset_t) list * answer_t

type run_t =
  | DoIt  of (res_t -> (res_t -> t) -> t)
  | GotIt of res_t

(* The table of running resolutions is keyed by [(name * rrt)].  Each
   entry contains a [run_state] with thread id, start time, list of
   mvars on which waiters are blocked and list of waiter's ids (only
   for display). *)

type run_state = {
    id : int;
    started_on : float;
    mutable mvars : run_t mvar list;
    mutable waiters : int list
  }

let running : (name * rrt, run_state) Hashtbl.t = Hashtbl.create 100;;


(* 

   See if a similar resolution is already running.  If so, we block
   until the answer is available (our continuation is added to the
   entry) and receive [GotIt result].  If not, we get [DoIt f], a
   handle to apply our result on when we'll have it.  This will give
   the answer to all the waiters and remove the entry from the table.

*)

let string_of_rslt n t =
  Printf.sprintf "%s %s" (string_of_name n) (string_of_rrt t)

let dbg_req n t id run =
  dbg (Printf.sprintf "R:= .%i. %s queuing on .%i. (%inth entry)"
	 id (string_of_rslt n t) run.id (List.length run.mvars))

let dbg_loop n t id rid =
  dbg (Printf.sprintf "R:! .%i. %s looping on .%i." id  (string_of_rslt n t) rid)

let check_running n t id pts k =
  try
    let run = Hashtbl.find running (n,t) in
    if List.mem run.id pts then (
      dbg_loop n t id run.id; k (GotIt([], RFailed "loop!!!")))
    else
      let mv = make_mvar () in
      run.mvars <- mv :: run.mvars;
      run.waiters <- id :: run.waiters;
      dbg_req n t id run;
      take_mvar mv >>= k
  with Not_found -> 
    dbg ("R:+ ." ^ string_of_int id ^ ". " ^  (string_of_rslt n t));
    let run = {
      started_on = Unix.gettimeofday();
      mvars = [];
      id = id;
      waiters = []
    }
    in
    Hashtbl.add running (n,t) run;

    let handle (cns,res) k =
      dbg (Printf.sprintf "R:- .%i. %s (%i waiters %s), %f ms" run.id (string_of_rslt n t) 
	     (List.length run.mvars)
	     (String.concat " " (List.map (Printf.sprintf ".%i.") run.waiters))
	     ((Unix.gettimeofday() -. run.started_on) *. 1000.));
      Hashtbl.remove running (n,t);
      let rec loop l () = 
	match l with
	| [] -> k (cns, res)
	| h::t -> put_mvar h (GotIt(cns,res)) >>= loop t
      in loop run.mvars ()
    in
    k (DoIt handle)

(* Dump functions to be used by the [mng] thread. *)

let dump_rslt_entry (n,t) =
  let run = Hashtbl.find running (n,t) in
  Printf.sprintf ".%i.:   %s, started on %s (%.2f ms ago), %i waiters [ %s ]" 
    run.id (string_of_rslt n t) (Misc.string_of_date run.started_on) 
    ((Unix.gettimeofday() -. run.started_on) *. 1000.)
    (List.length run.mvars) (String.concat " " (List.map (Printf.sprintf ".%i.") run.waiters))

let dump_rslt () =
  let keys = Hashtbl.fold (fun key v acc -> key :: acc) running []
  in
  String.concat "\n" (List.map (fun key -> dump_rslt_entry key) keys), List.length keys



(*i bug: the terminate was forgot + thread context is lost *)
(*
let check_running n t k =
  try
    let (_,l) = Hashtbl.find running (n,t) in
    l := k :: !l;
    dbg ("R:= " ^ (string_of_rslt n t) ^ " " ^ string_of_int (List.length !l) ^ "nth entry");
    terminate ()
  with Not_found -> 
    let l = ref [] in 
    dbg ("R:+ " ^  (string_of_rslt n t));
    Hashtbl.add running (n,t) (Unix.gettimeofday(), l);
    let handle r =
      let nb = List.length !l in
      let (d,_) = Hashtbl.find running (n,t) in
      dbg (Printf.sprintf "R:- %s (%i waiters), %f ms" (string_of_rslt n t) nb
	  ((Unix.gettimeofday() -. d) *. 1000.));
      Hashtbl.remove running (n,t);
      List.iter (fun k -> spawn (fun () -> k (GotIt r))) !l
    in
    k (DoIt handle)
i*)
